import sys
from galprime.isophote_l import Ellipse, EllipseGeometry
from numpy import max, pi, log, unravel_index, argmax, ceil
from photutils import data_properties
from tqdm import tqdm



def isophote_fitting(data, config=None, centre_method='standard'):
    """ Wrapper for photutils.isophote methods

    Generates a table of results from isophote fitting analysis. This uses photutils Isophote procedure, which is
    effectively IRAF's Ellipse() method.
    Iterates over many possible input ellipses to force a higher success rate.

    Args:
        data: The input cutout
        config: The configuration parameters, which contains details like STEP and LINEAR for profile extraction.
        centre_method: Which method to use to determine where to place the centre of the first ellipse.
            'standard': place at the exact centre of the cutout
            'max': place at the maximum pixel in the cutout

    Returns:
        The table of results, or an empty list if not fitted successfully.

    """
    # Set-up failsafe in case of strange infinte loops in photutils
    # warnings.filterwarnings("error")

    fail_count, max_fails = 0, 1000
    linear = False if config is None else config["LINEAR"]
    step = 1. if config is None else config["LINEAR_STEP"]
    verbose = False if config is None else config["VERBOSE"]

    # Get centre of image and cutout halfwidth
    if centre_method == 'standard':
        centre = (data.shape[0]/2, data.shape[1]/2)
    elif centre_method == 'max':
        centre = unravel_index(argmax(data), data.shape)
    else:
        centre = (data.shape[0] / 2, data.shape[1] / 2)

    cutout_halfwidth = max((ceil(data.shape[0] / 2), ceil(data.shape[1] / 2)))

    fitting_list = []

    # First, try obtaining morphological properties from the data and fit using that starting ellipse
    try:
        morph_cat = data_properties(log(data))
        r = 2.0
        pos = (morph_cat.xcentroid, morph_cat.ycentroid)
        a = morph_cat.semimajor_sigma.value * r
        b = morph_cat.semiminor_sigma.value * r
        theta = morph_cat.orientation.value

        geometry = EllipseGeometry(pos[0], pos[1], sma=a, eps=(1 - (b / a)), pa=theta)
        flux = Ellipse(data, geometry)
        fitting_list = flux.fit_image(maxit=100, maxsma=cutout_halfwidth, step=step, linear=linear,
                                      maxrit=cutout_halfwidth / 3)
        if len(fitting_list) > 0:
            return fitting_list

    except KeyboardInterrupt:
        sys.exit(1)
    except (RuntimeError, ValueError, OverflowError, IndexError):
        fail_count += 1
        if fail_count >= max_fails:
            return []

    # If that fails, test a parameter space of starting ellipses
    try:
        for angle in range(0, 180, 45):
            for sma in range(2, 26, 5):
                for eps in (0.3, 0.5, 0.9):
                    geometry = EllipseGeometry(float(centre[0]), float(centre[1]), eps=eps,
                                               sma=sma, pa=angle * pi / 180.)
                    flux = Ellipse(data, geometry)
                    fitting_list = flux.fit_image(maxsma=cutout_halfwidth, step=step, linear=linear,
                                                  maxrit=cutout_halfwidth / 3)
                    if len(fitting_list) > 0:
                        return fitting_list

    except KeyboardInterrupt:
        sys.exit(1)
    except (RuntimeError, ValueError, OverflowError, IndexError):

        # print("RuntimeError or ValueError")
        fail_count += 1
        if fail_count >= max_fails:
            return []
    except IndexError:
        fail_count += 1
        if fail_count >= max_fails:
            return []

    return fitting_list



from astropy.convolution import Gaussian2DKernel
from astropy.stats import gaussian_fwhm_to_sigma
import numpy as np
from photutils import detect_threshold, detect_sources, deblend_sources


class MaskedCutout:
    def __init__(self, cutout=None, mask=None, masked_cutout=None, config=None, 
                 segment_array=None, mask_info = None):
        self.cutout = cutout
        self.mask = mask
        self.segment_array = segment_array
        self.masked_cutout = masked_cutout
        self.mask_info = {} if mask_info is None else mask_info
        
        if config is not None:
            params = config["MASK_PARAMS"]
        else:
            params = [1., 2.0, 11]
        
        self.nsigma, self.gauss_width, self.npixels = params
        
    def gen_mask(self, nsigma=None, gauss_width=None, npixels=None, omit_centre=True, deblend=True, omit=None):
        nsigma = self.nsigma if nsigma is None else nsigma
        gauss_width = self.gauss_width if gauss_width is None else gauss_width
        npixels = self.npixels if npixels is None else npixels
        
        sigma = gauss_width * gaussian_fwhm_to_sigma
        kernel = Gaussian2DKernel(sigma)
        
        # First get our segment array (deblended or not)
        threshold = detect_threshold(self.cutout, nsigma=nsigma)
        segments = detect_sources(self.cutout, threshold, npixels=npixels, filter_kernel=kernel)
        
        if deblend:
            try:
                segments = deblend_sources(self.cutout, segments, npixels=npixels, filter_kernel=kernel)
            except ImportError:
                print("Skimage not working!")
        
        self.segment_array = segments
        
        # Now we apply the mask
        if omit is None:
            if omit_centre:
                segments_shape = self.segment_array.data.shape
                centre_value = self.segment_array.data[int(segments_shape[0]/2), int(segments_shape[1]/2)]
                omit = [0, centre_value]
            else:
                omit=[0]
        
        self.mask = np.isin(self.segment_array.data, omit, invert=True).astype(int)
        n_masked = np.sum(self.mask)

        self.mask_info["NSEG"] = self.segment_array.nlabels
        self.mask_info["CENT"] = centre_value
        self.mask_info["N_MASKED"] = n_masked
        self.mask_info["P_MASKED"] = n_masked / (self.cutout.shape[0] * self.cutout.shape[1])
        
    def apply_mask(self):
        self.masked_cutout = np.copy(self.cutout)
        self.masked_cutout[self.mask == 1] = np.nan
        return self.masked_cutout
    
    def mask_cutout(self, nsigma=None, gauss_width=None, npixels=None, omit_centre=True, deblend=True, omit=None):
        # If a mask has not been generated yet, generate the mask
        if self.mask is None:
            self.gen_mask(nsigma=nsigma, gauss_width=gauss_width, npixels=npixels, 
                     omit_centre=omit_centre, deblend=deblend, omit=omit)
        
        self.apply_mask()



""" Configuration settings methods.

This module contains all methods required to properly set up or read a configuration file for use with TBriDGE.

"""

import urllib.request
from numpy import arange


def load_config_file(filename, verbose_test=False):
    """ Loads in a config file for GalPrime to run

    :param filename: Filename (can be absolute or relative path, or a URL) to read config file from.
    :type filename: str
    :param verbose_test: If true, will create a helpful printout of params to ensure that the parameters are
        loading in properly.
    :type verbose_test: bool, optional
    :return: Configuration file as a dict object.
    :rtype: dict
    """
    config_values = {}

    # First try to open things locally. If that doesn't work try it as a URL
    try:
        config_lines = open(filename, "r").readlines()
    except FileNotFoundError:
        try:
            r = urllib.request.urlopen(filename)
            config_lines = []
            for line in r:
                config_lines.append(line.decode("utf-8"))
        except:
            print("Failed to get any file")
            return None

    for line in config_lines:
        line = line.strip()
        if len(line) == 0 or line[0] == "#":
            continue
        splits = line.split("=")
        config_values[splits[0].strip()] = splits[1].strip()

    for n in config_values:
        value = config_values[n]
        if value.lower() == "true":
            config_values[n] = True
            continue
        elif value.lower() == "false":
            config_values[n] = False
            continue

    # Adjust various parameters to their proper types (could be done though the code itself but this
    # ultimately feels cleaner to me in the long run
    config_values["VERBOSE"] = bool(config_values["VERBOSE"])
    config_values["DEBUG"] = bool(config_values["DEBUG"])

    config_values["SIZE"] = int(config_values["SIZE"])
    config_values["EXTRACTION_SIZE"] = int(config_values["EXTRACTION_SIZE"])
    config_values["CORES"] = int(config_values["CORES"])
    config_values["ZEROPOINT"] = int(config_values["ZEROPOINT"])
    config_values["ARC_CONV"] = float(config_values["ARC_CONV"])
    config_values["N_MODELS"] = int(config_values["N_MODELS"])
    config_values["N_BACKGROUNDS"] = int(config_values["N_BACKGROUNDS"])
    config_values["LINEAR_STEP"] = float(config_values["LINEAR_STEP"])
    config_values["TIME_LIMIT"] = float(config_values["TIME_LIMIT"])
    config_values["BOX_SIZE"] = int(config_values["BOX_SIZE"])
    config_values["FILTER_SIZE"] = int(config_values["FILTER_SIZE"])
    
    config_values["BIN_NAMES"] = [n.strip() for n in config_values["BIN_NAMES"].split(",")]
    
    # Turn all bins in numpy aranges (just to simplify the process). Will also add a x_step parameter
    for n in ("BIN_0", "BIN_1", "BIN_2"):
        if len(config_values[n]) == 0:
            config_values[n] = None
            continue
        value_string = config_values[n].split(",")
        bin_output = arange(float(value_string[0]), float(value_string[1]), float(value_string[2]))

        config_values[n] = bin_output
        config_values[n.split("_")[0] + "_STEP"] = float(value_string[2])

    value_string = config_values["MASK_PARAMS"].split(",")
    config_values["MASK_PARAMS"] = [float(value_string[0]), float(value_string[1]), int(value_string[2])]

    if verbose_test:
        for n in config_values:
            print(n, config_values[n], type(config_values[n]))

    return config_values


def print_config(config):
    """ Print out the configuration file values"""
    for n in config:
        print(n, "\t", config[n], "\t", type(config[n]))


def default_config_params():
    """
    Dumps a dict object containing all default parameters in proper type format.

    :return: Dictionary of config file values
    :rtype: dict
    """
    default_params = {
        "VERBOSE": True,
        "CATALOG": "cat.fits",
        "IMAGE_DIRECTORY": "images/",
        "PSF_FILENAME": "i_psfs.fits",
        "OUT_DIR": "out/",
        "BIN_NAMES": "BIN_1, BIN_2, BIN_3",

        "MAG_KEY": "i",
        "R50_KEY": "R50S",
        "N_KEY": "SERSIC_NS",
        "ELLIP_KEY": "ELLIPS",

        "N_MODELS": 100,
        "N_BACKGROUNDS": 25,

        "SIZE": 100,
        "EXTRACTION_SIZE": 100,
        "BAND": "i",
        "ZEROPOINT": 27,
        "CORES": 4,
        "ARC_CONV": 0.2,

        "BIN_0": arange(10., 12., 0.4),
        "BIN_1": arange(0.1, 0.9, 0.2),
        "BIN_2": arange(0.0, 1, 0.5),

        "LINEAR": True,
        "LINEAR_STEP": 1,
        "TIME_LIMIT": 60,
        
        "MASK_PARAMS": [1, 2.0, 11],
        "BOX_SIZE": 40,
        "FILTER_SIZE": 3
    }

    return default_params

# TODO I need to finish up adjusting these to the new corrected galprime file

def dump_default_config_file(directory=""):
    """ Dumps a default configuration file with all necessary parameters in the directory

    :param directory: Directory to write file to, defaults to local directory.
    :type directory: str, optional
    """
    lines = ["# Set verbosity printouts. VERBOSE: General printouts. TEST_VERBOSE: Additional printouts.",
             "VERBOSE             = True",
             "TEST_VERBOSE        = False",
             "",
             "# Directories and filenames -- Input and output",
             "# For SAVE CUTOUTS, options are 'none', 'mosaic', 'stitch', and 'fits'",
             "CATALOG             = cat.fits",
             "IMAGE_DIRECTORY     = images/",
             "PSF_FILENAME        = i_psfs.fits",
             "OUT_DIR             = out/",
             "SAVE_CUTOUTS        = none",
             "CUTOUT_FRACTION     = 0.2",

             "",
             "# Keys for masses, redshifts, and star-formation probability.",
             "MASS_KEY            = MASSES",
             "Z_KEY               = REDSHIFTS",
             "SFPROB_KEY          = SFPROBS",
             "",
             "# Keys for structural parameters. Magnitudes, half-light radii, Sersic index, ellipticity",
             "MAG_KEY             = i",
             "R50_KEY             = R50S",
             "N_KEY               = SERSIC_NS",
             "ELLIP_KEY           = ELLIPS",
             "",
             "# Cutout size, band, num-cores, arcsecs per pix, N models to gen, number of bgs to gen",
             "SIZE                = 100",
             "EXTRACTION_SIZE     = 100",
             "BAND                = i",
             "ZEROPOINT           = 27",
             "CORES               = 4",
             "ARC_CONV            = 0.2",
             "N_MODELS            = 100",
             "SAME_BGS            = True",
             "N_BGS               = 50",
             "",
             "# Bins to run through. (LOWER BOUND, UPPER BOUND, BIN WIDTH)",
             "# Note that the bins are defined by the LOWER BOUND to LOWER BOUND + BIN WIDTH",
             "MASS_BINS           = 10., 12., 0.4",
             "REDSHIFT_BINS       = 0.1, 0.9, 0.2",
             "SFPROB_BINS         = 0.0, 1, 0.5",
             "",
             "# Parameters for profile extraction.",
             "LINEAR              = True",
             "LINEAR_STEP         = 1",
             "USE_ALARM           = True",
             "ALARM_TIME          = 60",
             "",
             "# Parameters for Masking and background estimation ... NSIGMA, GAUSS_WIDTH, NPIX",
             "# Options for BG estimation: ellipse, circle, sigmaclip",
             "MASK_PARAMS         = 1, 2.0, 11",
             "BG_PARAMS           = ellipse",
             "BOX_SIZE            = 41",
             "FILTER_SIZE         = 3"
             ]

    with open(directory + "config.tbridge", "w+") as f:
        for n in lines:
            f.write(n + "\n")


def config_to_file(config, filename="config_out.galprime"):
    """ Write a config dict to a file.

    :param config: The configuration parameters.
    :type config: dict
    :param filename: The location where the parameters will be written, defaults to config_out.galprime.
    :type filename: str, optional
    """
    with open(filename, mode="w+") as f:
        for n in config:
            line = n + "\t" + str(config[n]) + "\t" + str(type(config[n])) + "\n"
            f.write(line)


def dump_default_config_file_koe(directory=""):
    """
    Dumps a default configuration file with all necessary parameters for using the KOE extraction pipeline.
    Args:
        directory: Directory to write file to. OPTIONAL.
    """
    lines = ["# Set verbosity printouts. VERBOSE: General printouts. TEST_MODE: Only extract 10 profiles as a check .",
             "VERBOSE             = True",
             "TEST_MODE        = False",
             "",
             "# Directories and filenames -- Input and output",
             "CATALOG             = cat.fits",
             "IMAGE_DIRECTORY     = images/",
             "OUT_DIR             = out/",
             "",
             "# Keys for RA, DEC, and Z",
             "RA                  = RA:",
             "DEC                 = DEC",
             "",
             "# Cutout size, band, num-cores, arcseconds per pix",
             "SIZE                = 100",
             "BAND                = i",
             "ZEROPOINT           = 27",
             "CORES               = 4",
             "ARC_CONV            = 0.2",
             "",
             "# Parameters for profile extraction.",
             "LINEAR              = True",
             "LINEAR_STEP         = 1",
             "USE_ALARM           = True",
             "ALARM_TIME          = 60",
             "",
             "# Parameters for Masking and background estimation ... NSIGMA, GAUSS_WIDTH, NPIX",
             "# Options for BG estimation: ellipse, circle, sigmaclip, 2D",
             "MASK_PARAMS         = 1, 2.0, 11",
             "BG_PARAMS           = sigmaclip",
             "BOX_SIZE            = 41",
             "FILTER_SIZE         = 6"
             ]

    with open(directory + "koe_config.tbridge", "w+") as f:
        for n in lines:
            f.write(n + "\n")


def load_config_file_koe(filename):
    """ Loads in a config file for KOE to run

    Args:
        filename: Filename (can be absolute or relative path, or a URL) to read config file from.
    Returns:
        dict: Configuration file as a dict object.
    """
    config_values = {}

    # First try to open things locally. If that doesn't work try it as a URL
    try:
        config_lines = open(filename, "r").readlines()
    except FileNotFoundError:
        try:
            r = urllib.request.urlopen(filename)
            config_lines = []
            for line in r:
                config_lines.append(line.decode("utf-8"))
        except:
            print("Failed to get any file")
            return None

    for line in config_lines:
        line = line.strip()
        if len(line) == 0 or line[0] == "#":
            continue
        splits = line.split("=")
        config_values[splits[0].strip()] = splits[1].strip()

    for n in config_values:
        value = config_values[n]
        if value.lower() == "true":
            config_values[n] = True
            continue
        elif value.lower() == "false":
            config_values[n] = False
            continue
    config_values["SIZE"] = int(config_values["SIZE"])
    config_values["CORES"] = int(config_values["CORES"])
    config_values["ZEROPOINT"] = int(config_values["ZEROPOINT"])
    config_values["ARC_CONV"] = float(config_values["ARC_CONV"])
    config_values["LINEAR_STEP"] = float(config_values["LINEAR_STEP"])
    config_values["ALARM_TIME"] = int(config_values["ALARM_TIME"])
    config_values["BOX_SIZE"] = int(config_values["BOX_SIZE"])
    config_values["FILTER_SIZE"] = int(config_values["FILTER_SIZE"])

    value_string = config_values["MASK_PARAMS"].split(",")
    config_values["MASK_PARAMS"] = [float(value_string[0]), float(value_string[1]), int(value_string[2])]

    return config_values



""" Background estimation and subtraction module.

This module contains all methods available to estimate the backgrounds of cutouts,
and to subtract these backgrounds from cutouts or profiles as a means to correct them.

"""

import numpy as np
from astropy.stats import sigma_clipped_stats, SigmaClip
from astropy.convolution import convolve, Tophat2DKernel

from photutils import Background2D, MedianBackground, BkgZoomInterpolator, make_source_mask


def background_2D(img, mask, box_size, interp=None, filter_size=1,
                  exclude_percentile=90):
    """ Run photutils background with SigmaClip and MedianBackground

    :param img: The 2D input image
    :type img: array_like (np.ndarray)
    :param mask: The 2D image mask
    :type mask: array_like (np.ndarray)
    :param box_size: The size of the box used in the 2D background. It should be larger than
        the largest objects in your image but still sufficiently small to capture large-scale
        structure.
    :type box_size: int, optional
    :param filter_size: The window size of the median filter being applied to the background image. A
        higher filter size will result in more smoothing to the background.
    :type filter_size: int, optional
    :param exclude_percentile: If the percentage of masked pixels in a box is above the exclude percentile,
        it is not included in determining the 2D background.
    :type exclude_percentile: float, optional
    """
    if interp is None:
        interp = BkgZoomInterpolator()
    return Background2D(img, box_size,
                        sigma_clip=SigmaClip(sigma=3.),
                        filter_size=filter_size,
                        bkg_estimator=MedianBackground(),
                        exclude_percentile=exclude_percentile,
                        mask=mask,
                        interpolator=interp)


class SourceMask:
    def __init__(self, img, nsigma=3., npixels=10, mask=None):
        """ Helper for making & dilating a source mask.
             See Photutils docs for make_source_mask.

            :param img: The image that is to be masked
            :type img: array_like (np.ndarray)
            :param nsigma: The sigma detection threshold for the source mask, defaults to 3
            :type nsigma: float, optional
            :param npixels: The number of required pixels for a detection, defaults to 10
            :type npixels: int, optional
            :param mask: An already-suppled mask for objects in the image.
            :type array_like (np.ndarray), optional
        """
        self.img = img
        self.nsigma = nsigma
        self.npixels = npixels
        if mask is None:
            self.mask = np.zeros(self.img.shape, dtype=np.bool)
        else:
            self.mask = mask

    def single(self, filter_fwhm=3., tophat_size=5., mask=None):
        """ Mask on a single scale """
        if mask is None:
            image = self.img
        else:
            image = self.img * (1 - mask)
        mask = make_source_mask(image, nsigma=self.nsigma,
                                npixels=self.npixels,
                                dilate_size=1, filter_fwhm=filter_fwhm)
        return dilate_mask(mask, tophat_size)

    def multiple(self, filter_fwhm=[3.], tophat_size=[3.], mask=None):
        """ Mask repeatedly on different scales """
        if mask is None:
            self.mask = np.zeros(self.img.shape, dtype=np.bool)
        for fwhm, tophat in zip(filter_fwhm, tophat_size):
            smask = self.single(filter_fwhm=fwhm, tophat_size=tophat)
            self.mask = self.mask | smask  # Or the masks at each iteration

        return self.mask

    def dilated(self, tophat_size=[3.], mask=None):
        """Mask using simple dilation"""
        if mask is None:
            self.mask = self.single()
        for tophat in tophat_size:
            smask = dilate_mask(self.mask, tophat)
            self.mask = self.mask | smask  # Or the masks at each iteration

        return self.mask


def dilate_mask(mask, tophat_size):
    """ Dilate a mask with a tophat kernel. """
    area = np.pi * tophat_size ** 2.
    kernel = Tophat2DKernel(tophat_size)
    dilated_mask = convolve(mask, kernel) >= 1. / area
    return dilated_mask


def estimate_background_sigclip(cutout, config=None, nsigma=2, npixels=3, dilate_size=7):
    """ Estimate the background mean, median, and standard deviation of a cutout using sigma-clipped-stats """

    nsigma is config["MASK_PARAMS"][0] if config is not None else nsigma
    npixels is config["MASK_PARAMS"][2] if config is not None else npixels

    bg_mask = make_source_mask(cutout, nsigma=nsigma, npixels=npixels, dilate_size=dilate_size)

    bg_mean, bg_median, bg_std = sigma_clipped_stats(cutout, sigma=3.0, mask=bg_mask)

    return bg_mean, bg_median, bg_std




import galprime
import numpy as np
import os

from matplotlib import pyplot as plt

from astropy.io import fits
from astropy.convolution import Gaussian2DKernel

from astropy.visualization import ZScaleInterval

from numpy.random import randint
from scipy.signal import convolve2d

import galprime

from pebble import ProcessPool
from concurrent.futures import TimeoutError

class GalPrimeError(Exception):
    pass


class GPrime():
    """ Class to create and run GalPrime simulations """
    
    def __init__(self, config=None, psfs=None, backgrounds=None, mag_kde=None):

        if config is None:
            self.config = galprime.default_config_params()
        elif type(config) == str:
            self.config = galprime.load_config_file(config)
        elif type(config) == dict:
            self.config = config
        else:
            raise GalPrimeError("Uncertain format for config file. File needs to be dict object or str filename.")
            sys.exit(1)
        
        self.binned_objects = None
        self.psfs = psfs
        self.backgrounds = backgrounds
        self.mag_kde = mag_kde

    def load_bins(self):
        self.binned_objects = galprime.bin_catalog(self.config)
    
    
    def pipeline(self, max_bins=None, mag_kde=None, process_method=None,
                progress_bar=False, debug=False, 
                 table_names=["ID","IMG", "X", "Y", "RA", "DEC", "MAGS", "R50S", "NS", "ELLIPS", "MASS_MED",
                              "ZPHOT", "sfProb", "I_R50", "PA", "R50_PIX", "BG_MEAN", "BG_MED", "BG_STD",
                              "NSEG_BGA", "CENT_BGA", "N_MASKED_BGA", "P_MASKED_BGA", "NSEG_BGSUB", "CENT_BGSUB", 
                              "N_MASKED_BGSUB", "P_MASKED_BGSUB"]):
        
        # Bin out catalog based on the catalog in the configuration file
        binned_objects = galprime.bin_catalog(self.config)
        max_bins = len(binned_objects.bins) if max_bins is None else max_bins
        
        verbose = self.config["VERBOSE"]

        if not os.path.isdir(self.config["OUT_DIR"]):
            os.mkdir(self.config["OUT_DIR"])
        
        # Run through bins and process using the method
        for i in range(max_bins):
            current_bin = binned_objects.bins[i]
            
            columns = current_bin.columns()
            
            # Generate the object KDE that will be used by this bin
            kde = galprime.object_kde(columns)
            
            # Create our container object (this is what will be multithreaded) ###
            bg_index = randint(len(self.backgrounds.cutouts))
            
            job_list, job_results = [], []
            # Set up the job pool
            print("Generating", self.config["N_MODELS"], "models for bin:", current_bin.bin_params)
            with ProcessPool(max_workers=self.config["CORES"]) as pool:
                for i in range(self.config["N_MODELS"]):
                    job_list.append(pool.schedule(process_method,
                                                args=(self, current_bin, kde),
                                                timeout=self.config["TIME_LIMIT"]))
            # Now get all the results
            for i, job in enumerate(job_list):
                try:
                    job_results.append(job.result())
                except Exception as error:
                    print(error.args, i)
            
            print(len(job_results))



class GalPrimeContainer:
    
    def __init__(self, config=None, kde=None, mag_kde=None, background=None, psf=None,
                 model_cutout=None, background_cutout=None, bg_model=None,
                 bgadded_masked=None, bgsub_cutout=None, bgsub_masked=None, 
                 model_profile=None, bgadded_profile=None, bgsub_profile=None,
                 metadata=None):
        
        if config == None:
            self.config = galprime.default_config_params()
        else:
            self.config = config
        
        if metadata is None:
            self.metadata = {}
        else:
            self.metadata = metadata
            
        # Generate Container ID
        self.metadata["ID"] = np.random.randint(1e15,1e16)
        
        self.kde = kde
        self.mag_kde = mag_kde
        self.background = background
        
        # Models without masking
        self.model = model_cutout
        self.convolved_model = None
        self.background_cutout = background_cutout
        self.bgsub_cutout = bgsub_cutout
        self.psf = psf
        
        # Background-added cutouts after masking
        self.bgadded_masked = bgadded_masked
        self.bgsub_masked = bgsub_masked
        self.bg_model = bg_model
        
        # Extracted profiles
        self.model_profile = model_profile
        self.bgadded_profile = bgadded_profile
        self.bgsub_profile = bgsub_profile
    
    # Functions to run ########################################################################
    def generate_model(self):
        if self.config["DEBUG"]:
            print("Generating Sersic model cutout of size", self.config["EXTRACTION_SIZE"])    
        self.model, model_params = galprime.model_from_kde(self.kde, self.config, self.mag_kde, 
                                                           names=self.metadata["NAMES"])
        self.metadata.update(model_params)
        
        
    def get_background(self):
        # Get background and convolve model with associated Best PSF
        
        # self.psf = galprime.get_closest_psf(psfs, self.metadata["RA"], self.metadata["DEC"])
                
        # Convolve model with the PSF, and add the model to the background
        self.convolved_model = convolve2d(self.model, self.psf, mode='same')
        self.bgadded_cutout = self.convolved_model + self.background_cutout
        
        bg_stats = galprime.estimate_background_sigclip(self.bgadded_cutout, config=self.config)
        self.metadata["BG_MEAN"], self.metadata["BG_MED"], self.metadata["BG_STD"] = bg_stats
    
    def subtract_background(self):
        if self.config["DEBUG"]:
            print("Subtracting the background from the bgadded cutout")
        sm = galprime.SourceMask(self.convolved_model + self.background_cutout) 
        mask = sm.multiple(filter_fwhm=[1,3,5], tophat_size=[4,2,1])
        self.bgmask = mask
        
        bg_2D = galprime.background_2D(self.bgadded_cutout, self.bgmask, self.config["BOX_SIZE"], filter_size=self.config["FILTER_SIZE"])
        self.bg_model = bg_2D.background
        
    def mask_cutouts(self):
        if self.config["DEBUG"]:
            print("Masking all profiles")
        # Apply masking to both bgadded and bgsub, add these to the container, and update the metadata
        bgadded_masked = galprime.MaskedCutout(cutout=self.bgadded_cutout, config=self.config)
        bgadded_masked.mask_cutout()
        self.bgadded_masked = bgadded_masked.masked_cutout
        self.metadata.update(galprime.dict_extend(bgadded_masked.mask_info, extension="_BGA"))

        bgsub_masked = galprime.MaskedCutout(cutout=self.bgadded_cutout - self.bg_model, config=self.config)
        bgsub_masked.mask_cutout()
        self.metadata.update(galprime.dict_extend(bgsub_masked.mask_info, extension="_BGSUB"))            
        self.bgsub_masked = bgsub_masked.masked_cutout

            
    def extract_profiles(self):
        if self.config["DEBUG"]:
            print("Extracting all profiles")
        self.model_profile = galprime.isophote_fitting(self.convolved_model, config=self.config)
        self.bgadded_profile = galprime.isophote_fitting(self.bgadded_masked, config=self.config)
        self.bgsub_profile = galprime.isophote_fitting(self.bgsub_masked, config=self.config)
            
            
    def save_data(self):
        if self.config["DEBUG"]:
            print("Saving data")
    
    def plot_data(self, filename="out.png"):
        fig, ax = plt.subplots(2,5, facecolor="white")
        fig.set_figheight(5)
        fig.set_figwidth(12)
        
        scale = ZScaleInterval(contrast=0.5).get_limits(self.bgadded_cutout)
        
        ax[0][0].imshow(self.model, origin='lower', vmin=scale[0], vmax=scale[1], cmap="Greys")
        ax[0][1].imshow(self.convolved_model, origin='lower', vmin=scale[0], vmax=scale[1], cmap="Greys")
        ax[0][2].imshow(self.psf, origin='lower', cmap="Greys")
        ax[0][3].imshow(self.background_cutout, origin='lower', vmin=scale[0], vmax=scale[1], cmap="Greys")
        ax[0][4].imshow(self.convolved_model + self.background_cutout, origin='lower', vmin=scale[0], 
        vmax=scale[1], cmap="Greys")
        
        ax[1][0].imshow(self.bg_model, origin="lower", vmin=scale[0], vmax=scale[1], cmap="Greys")
        ax[1][1].imshow(self.bgmask, origin='lower', cmap="Greys")
        ax[1][2].imshow(self.bgadded_masked, origin='lower', vmin=scale[0], vmax=scale[1], cmap="Greys")
        ax[1][3].imshow(self.bgsub_masked, origin='lower', vmin=scale[0], vmax=scale[1], cmap="Greys")
        
        
        ax[0][0].set_title("Model")
        ax[0][1].set_title("Convolved Model")
        ax[0][2].set_title("PSF")
        ax[0][3].set_title("Background Cutout")
        ax[0][4].set_title("Combined Cutout")
        
        ax[1][0].set_title("Model Background")
        ax[1][1].set_title("Background Mask")
        ax[1][2].set_title("Masked BGAdded")
        ax[1][3].set_title("Masked BG-Subbed")
        ax[1][4].set_title("Profiles: Flux vs " + r'$R^{1/4}$')
        
        bare_prof = self.model_profile
        bgadded_prof = self.bgadded_profile
        bgsub_prof = self.bgsub_profile
        
        
        ax[1][4].plot(bare_prof.sma ** (1/4), np.log10(bare_prof.intens), color="red")
        ax[1][4].plot(bgadded_prof.sma ** (1/4), np.log10(bgadded_prof.intens), color="blue")
        ax[1][4].plot(bgsub_prof.sma ** (1/4), np.log10(bgsub_prof.intens), color="gold")

        for j in range(5):
            for i in range(2):
                ax[i][j].set_xticks([])
                ax[i][j].set_yticks([])
        
        ax[1][4].set_xticks(np.arange(0, np.max(bare_prof.sma ** (1/4))))
        
        plt.suptitle("ID: " + str(self.metadata["ID"]))
        plt.tight_layout()
        if filename is None:
            plt.show()
        else:
            plt.savefig(filename, dpi=150)
        
    
    def save_outputs(self, filename="GalPrimeContainer.fits", overwrite=True):
        print("Saving output")
        out_HDUList = fits.HDUList()
        head = fits.Header()
        
        for n in self.metadata:
            val = self.metadata[n]
            if type(val) == list:
                continue
            head[n] = self.metadata[n]
        
        out_HDUList.append(fits.PrimaryHDU(header=head))
        out_HDUList.append(fits.ImageHDU(data=self.convolved_model, name="MODEL"))
        out_HDUList.append(fits.ImageHDU(data=self.bgadded_cutout, name="BGA_IMG"))
        out_HDUList.append(fits.ImageHDU(data=self.bgsub_cutout, name="BGS_IMG"))
        out_HDUList.append(fits.ImageHDU(data=self.bgadded_masked, name="BGAMASKED"))
        out_HDUList.append(fits.ImageHDU(data=self.bgsub_masked, name="BGSMASKED"))
        out_HDUList.append(fits.ImageHDU(data=self.bg_model, name="BGMODEL"))
        out_HDUList.append(fits.BinTableHDU(data=galprime.table_from_isolist(self.model_profile), name="MOD_PROF"))
        out_HDUList.append(fits.BinTableHDU(data=galprime.table_from_isolist(self.bgadded_profile), name="BGA_PROF"))
        out_HDUList.append(fits.BinTableHDU(data=galprime.table_from_isolist(self.bgsub_profile), name="BGS_PROF"))
       
        print(len(out_HDUList), type(out_HDUList))
        
        out_HDUList.writeto(filename, overwrite=overwrite)
        

    def process_object(self, plot=False):
        self.generate_model()
        self.get_background()
        self.subtract_background()
        self.mask_cutouts()
        self.extract_profiles()
        self.save_data()
        
        if plot:
            outdir = self.config["OUT_DIR"] + "pngs/"
            if not os.path.isdir(outdir):
                os.mkdir(outdir)
            self.plot_data(filename=outdir + "obj_" + str(self.metadata["ID"]) + ".png")


def gprime_single(gprime_obj, current_bin, kde):
    metadata = {}
    metadata["NAMES"] = current_bin.object_column_names
    
    bg_index = randint(len(gprime_obj.backgrounds.cutouts))
    metadata["BG_INDEX"] = bg_index
    
    metadata.update(gprime_obj.backgrounds.cutout_data[bg_index])

    # Get the best PSF is the user has supplied PSFs to the pipeline
    # Otherwise, just use a simple Gaussian2D kernel from astropy
    if gprime_obj.psfs is None:
        psf = Gaussian2DKernel(x_stddev=1.5, y_stddev=1.5)
    else:
        psf = galprime.get_closest_psf(gprime_obj.psfs, metadata["RA"], metadata["DEC"])

    # Generate our container
    container = GalPrimeContainer(config=gprime_obj.config, kde=kde, psf=psf,
                                    background_cutout=gprime_obj.backgrounds.cutouts[bg_index], 
                                    metadata=metadata)
    container.process_object(plot=True)
    return container



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
This module provides a class to fit ellipses.
"""

import math

from astropy import log
import numpy as np
import numpy.ma as ma

from .harmonics import (first_and_second_harmonic_function,
                        fit_first_and_second_harmonics)
from .isophote import CentralPixel, Isophote
from .sample import EllipseSample

__all__ = ['EllipseFitter']

__doctest_skip__ = ['EllipseFitter.fit']


PI2 = np.pi / 2
MAX_EPS = 0.95
MIN_EPS = 0.05

DEFAULT_CONVERGENCE = 0.05
DEFAULT_MINIT = 10
DEFAULT_MAXIT = 50
DEFAULT_FFLAG = 0.7
DEFAULT_MAXGERR = 0.5


class EllipseFitter:
    """
    Class to fit ellipses.

    Parameters
    ----------
    sample : `~photutils.isophote.EllipseSample` instance
        The sample data to be fitted.
    """

    def __init__(self, sample):
        self._sample = sample

    def fit(self, conver=DEFAULT_CONVERGENCE, minit=DEFAULT_MINIT,
            maxit=DEFAULT_MAXIT, fflag=DEFAULT_FFLAG, maxgerr=DEFAULT_MAXGERR,
            going_inwards=False):
        """
        Fit an elliptical isophote.

        Parameters
        ----------
        conver : float, optional
            The main convergence criterion. Iterations stop when the
            largest harmonic amplitude becomes smaller (in absolute
            value) than ``conver`` times the harmonic fit rms.  The
            default is 0.05.
        minit : int, optional
            The minimum number of iterations to perform. A minimum of 10
            (the default) iterations guarantees that, on average, 2
            iterations will be available for fitting each independent
            parameter (the four harmonic amplitudes and the intensity
            level). For the first isophote, the minimum number of
            iterations is 2 * ``minit`` to ensure that, even departing
            from not-so-good initial values, the algorithm has a better
            chance to converge to a sensible solution.
        maxit : int, optional
            The maximum number of iterations to perform.  The default is
            50.
        fflag : float, optional
            The acceptable fraction of flagged data points in the
            sample.  If the actual fraction of valid data points is
            smaller than this, the iterations will stop and the current
            `~photutils.isophote.Isophote` will be returned.  Flagged
            data points are points that either lie outside the image
            frame, are masked, or were rejected by sigma-clipping.  The
            default is 0.7.
        maxgerr : float, optional
            The maximum acceptable relative error in the local radial
            intensity gradient. This is the main control for preventing
            ellipses to grow to regions of too low signal-to-noise
            ratio.  It specifies the maximum acceptable relative error
            in the local radial intensity gradient.  `Busko (1996; ASPC
            101, 139)
            <https://ui.adsabs.harvard.edu/abs/1996ASPC..101..139B/abstract>`_
            showed that the fitting precision relates to that relative
            error.  The usual behavior of the gradient relative error is
            to increase with semimajor axis, being larger in outer,
            fainter regions of a galaxy image.  In the current
            implementation, the ``maxgerr`` criterion is triggered only
            when two consecutive isophotes exceed the value specified by
            the parameter. This prevents premature stopping caused by
            contamination such as stars and HII regions.

            A number of actions may happen when the gradient error
            exceeds ``maxgerr`` (or becomes non-significant and is set
            to `None`).  If the maximum semimajor axis specified by
            ``maxsma`` is set to `None`, semimajor axis growth is
            stopped and the algorithm proceeds inwards to the galaxy
            center. If ``maxsma`` is set to some finite value, and this
            value is larger than the current semimajor axis length, the
            algorithm enters non-iterative mode and proceeds outwards
            until reaching ``maxsma``.  The default is 0.5.
        going_inwards : bool, optional
            Parameter to define the sense of SMA growth. When fitting
            just one isophote, this parameter is used only by the code
            that defines the details of how elliptical arc segments
            ("sectors") are extracted from the image, when using area
            extraction modes (see the ``integrmode`` parameter in the
            `~photutils.isophote.EllipseSample` class).  The default is
            `False`.

        Returns
        -------
        result : `~photutils.isophote.Isophote` instance
            The fitted isophote, which also contains fit status
            information.

        Examples
        --------
        >>> from photutils.isophote import EllipseSample, EllipseFitter
        >>> sample = EllipseSample(data, sma=10.)
        >>> fitter = EllipseFitter(sample)
        >>> isophote = fitter.fit()
        """

        sample = self._sample

        # this flag signals that limiting gradient error (`maxgerr`)
        # wasn't exceeded yet.
        lexceed = False

        # these values count the number of fitting errors (to catch an infinite loop in interations) and
        # specify a max threshold
        iter_fails, max_iter_fails = 0, 100

        # here we keep track of the sample that caused the minimum harmonic
        # amplitude(in absolute value). This will eventually be used to
        # build the resulting Isophote in cases where iterations run to
        # the maximum allowed (maxit), or the maximum number of flagged
        # data points (fflag) is reached.
        minimum_amplitude_value = np.Inf
        minimum_amplitude_sample = None

        # these must be passed throughout the execution chain.
        fixed_parameters = self._sample.geometry.fix

        for i in range(maxit):
            # Force the sample to compute its gradient and associated values.
            sample.update(fixed_parameters)

            # The extract() method returns sampled values as a 2-d numpy array
            # with the following structure:
            # values[0] = 1-d array with angles
            # values[1] = 1-d array with radii
            # values[2] = 1-d array with intensity
            values = sample.extract()

            # Fit harmonic coefficients. Failure in fitting is
            # a fatal error; terminate immediately with sample
            # marked as invalid.
            try:
                coeffs = fit_first_and_second_harmonics(values[0], values[2])
                coeffs = coeffs[0]
            except Exception as e:
                iter_fails += 1
                if iter_fails >= max_iter_fails:
                    raise RuntimeError

                sample.geometry.fix = fixed_parameters
                return Isophote(sample, i + 1, False, 3)

            # Mask out coefficients that control fixed ellipse parameters.
            free_coeffs = ma.masked_array(coeffs[1:], mask=fixed_parameters)

            # Largest non-masked harmonic in absolute value drives the correction.
            largest_harmonic_index = np.argmax(np.abs(free_coeffs))
            largest_harmonic = free_coeffs[largest_harmonic_index]

            # see if the amplitude decreased; if yes, keep the
            # corresponding sample for eventual later use.
            if abs(largest_harmonic) < minimum_amplitude_value:
                minimum_amplitude_value = abs(largest_harmonic)
                minimum_amplitude_sample = sample

            # check if converged
            model = first_and_second_harmonic_function(values[0], coeffs)
            residual = values[2] - model

            if ((conver * sample.sector_area * np.std(residual))
                    > np.abs(largest_harmonic)):
                # Got a valid solution. But before returning, ensure
                # that a minimum of iterations has run.
                if i >= minit - 1:
                    sample.update(fixed_parameters)
                    return Isophote(sample, i + 1, True, 0)

            # it may not have converged yet, but the sample contains too
            # many invalid data points: return.
            if sample.actual_points < (sample.total_points * fflag):
                # when too many data points were flagged, return the
                # best fit sample instead of the current one.
                minimum_amplitude_sample.update(fixed_parameters)
                return Isophote(minimum_amplitude_sample, i + 1, True, 1)

            # pick appropriate corrector code.
            corrector = _CORRECTORS[largest_harmonic_index]

            # generate *NEW* EllipseSample instance with corrected
            # parameter.  Note that this instance is still devoid of other
            # information besides its geometry.  It needs to be explicitly
            # updated for computations to proceed.  We have to build a new
            # EllipseSample instance every time because of the lazy
            # extraction process used by EllipseSample code. To minimize
            # the number of calls to the area integrators, we pay a
            # (hopefully smaller) price here, by having multiple calls to
            # the EllipseSample constructor.
            sample = corrector.correct(sample, largest_harmonic)
            sample.update(fixed_parameters)

            # see if any abnormal (or unusual) conditions warrant
            # the change to non-iterative mode, or go-inwards mode.
            proceed, lexceed = self._check_conditions(
                sample, maxgerr, going_inwards, lexceed)

            if not proceed:
                sample.update(fixed_parameters)
                return Isophote(sample, i + 1, True, -1)

        # Got to the maximum number of iterations. Return with
        # code 2, and handle it as a valid isophote. Use the
        # best fit sample instead of the current one.
        minimum_amplitude_sample.update(fixed_parameters)
        return Isophote(minimum_amplitude_sample, maxit, True, 2)

    @staticmethod
    def _check_conditions(sample, maxgerr, going_inwards, lexceed):
        proceed = True

        # If center wandered more than allowed, put it back
        # in place and signal the end of iterative mode.
        # if wander:
        #     if abs(dx) > WANDER(al)) or abs(dy) > WANDER(al):
        #         sample.geometry.x0 -= dx
        #         sample.geometry.y0 -= dy
        #         STOP(al) = ST_NONITERATE
        #         proceed = False

        # check if an acceptable gradient value could be computed.
        if sample.gradient_error:
            if not going_inwards and (
                    sample.gradient_relative_error > maxgerr
                    or sample.gradient >= 0.0):
                if lexceed:
                    proceed = False
                else:
                    lexceed = True
        else:
            proceed = False

        # check if ellipse geometry diverged.
        if abs(sample.geometry.eps > MAX_EPS):
            proceed = False
        if (sample.geometry.x0 < 1. or
                sample.geometry.x0 > sample.image.shape[1] or
                sample.geometry.y0 < 1. or
                sample.geometry.y0 > sample.image.shape[0]):
            proceed = False

        # See if eps == 0 (round isophote) was crossed.
        # If so, fix it but still proceed
        if sample.geometry.eps < 0.:
            sample.geometry.eps = min(-sample.geometry.eps, MAX_EPS)
            if sample.geometry.pa < PI2:
                sample.geometry.pa += PI2
            else:
                sample.geometry.pa -= PI2

        # If ellipse is an exact circle, computations will diverge.
        # Make it slightly flat, but still proceed
        if sample.geometry.eps == 0.0:
            sample.geometry.eps = MIN_EPS

        return proceed, lexceed


class _ParameterCorrector:

    def correct(self, sample, harmonic):
        raise NotImplementedError


class _PositionCorrector(_ParameterCorrector):

    @staticmethod
    def finalize_correction(dx, dy, sample):
        new_x0 = sample.geometry.x0 + dx
        new_y0 = sample.geometry.y0 + dy

        return EllipseSample(sample.image, sample.geometry.sma, x0=new_x0,
                             y0=new_y0, astep=sample.geometry.astep,
                             sclip=sample.sclip, nclip=sample.nclip,
                             eps=sample.geometry.eps,
                             position_angle=sample.geometry.pa,
                             linear_growth=sample.geometry.linear_growth,
                             integrmode=sample.integrmode)


class _PositionCorrector0(_PositionCorrector):

    def correct(self, sample, harmonic):
        aux = -harmonic * (1. - sample.geometry.eps) / sample.gradient

        dx = -aux * math.sin(sample.geometry.pa)
        dy = aux * math.cos(sample.geometry.pa)

        return self.finalize_correction(dx, dy, sample)


class _PositionCorrector1(_PositionCorrector):

    def correct(self, sample, harmonic):
        aux = -harmonic / sample.gradient

        dx = aux * math.cos(sample.geometry.pa)
        dy = aux * math.sin(sample.geometry.pa)

        return self.finalize_correction(dx, dy, sample)


class _AngleCorrector(_ParameterCorrector):

    def correct(self, sample, harmonic):
        eps = sample.geometry.eps
        sma = sample.geometry.sma
        gradient = sample.gradient

        correction = (harmonic * 2. * (1. - eps) / sma / gradient /
                      ((1. - eps)**2 - 1.))

        # '% np.pi' to make angle lie between 0 and np.pi radians
        new_pa = (sample.geometry.pa + correction) % np.pi

        return EllipseSample(sample.image, sample.geometry.sma,
                             x0=sample.geometry.x0, y0=sample.geometry.y0,
                             astep=sample.geometry.astep, sclip=sample.sclip,
                             nclip=sample.nclip, eps=sample.geometry.eps,
                             position_angle=new_pa,
                             linear_growth=sample.geometry.linear_growth,
                             integrmode=sample.integrmode)


class _EllipticityCorrector(_ParameterCorrector):

    def correct(self, sample, harmonic):
        eps = sample.geometry.eps
        sma = sample.geometry.sma
        gradient = sample.gradient

        correction = harmonic * 2. * (1. - eps) / sma / gradient

        new_eps = min((sample.geometry.eps - correction), MAX_EPS)

        return EllipseSample(sample.image, sample.geometry.sma,
                             x0=sample.geometry.x0, y0=sample.geometry.y0,
                             astep=sample.geometry.astep, sclip=sample.sclip,
                             nclip=sample.nclip, eps=new_eps,
                             position_angle=sample.geometry.pa,
                             linear_growth=sample.geometry.linear_growth,
                             integrmode=sample.integrmode)


# instances of corrector code live here:
_CORRECTORS = [_PositionCorrector0(), _PositionCorrector1(),
               _AngleCorrector(), _EllipticityCorrector()]


class CentralEllipseFitter(EllipseFitter):
    """
    A special Fitter class to handle the case of the central pixel in
    the galaxy image.
    """

    def fit(self, conver=DEFAULT_CONVERGENCE, minit=DEFAULT_MINIT,
            maxit=DEFAULT_MAXIT, fflag=DEFAULT_FFLAG, maxgerr=DEFAULT_MAXGERR,
            going_inwards=False):
        """
        Perform just a simple 1-pixel extraction at the current (x0, y0)
        position using bilinear interpolation.

        The input parameters are ignored, but included simple to match
        the calling signature of the parent class.

        Returns
        -------
        result : `~photutils.isophote.CentralEllipsePixel` instance
            The central pixel value.  For convenience, the
            `~photutils.isophote.CentralEllipsePixel` class inherits
            from the `~photutils.isophote.Isophote` class, although it's
            not really a true isophote but just a single intensity value
            at the central position.  Thus, most of its attributes are
            hardcoded to `None` or other default value when appropriate.
        """
        # default values
        fixed_parameters = np.array([False, False, False, False])

        self._sample.update(fixed_parameters)
        return CentralPixel(self._sample)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
This module provides classes to store the results of isophote fits.
"""

from collections import OrderedDict

from astropy.table import QTable
import astropy.units as u
import numpy as np

from .harmonics import (first_and_second_harmonic_function,
                        fit_first_and_second_harmonics, fit_upper_harmonic)

__all__ = ['Isophote', 'IsophoteList']


class Isophote:
    """
    Container class to store the results of single isophote fit.

    The extracted data sample at the given isophote (sampled intensities
    along the elliptical path on the image) is also kept as an attribute
    of this class. The container concept helps in segregating
    information directly related to the sample, from information that
    more closely relates to the fitting process, such as status codes,
    errors for isophote parameters, and the like.

    Parameters
    ----------
    sample : `~photutils.isophote.EllipseSample` instance
        The sample information.
    niter : int
        The number of iterations used to fit the isophote.
    valid : bool
        The status of the fitting operation.
    stop_code : int
        The fitting stop code:

           *  0: Normal.
           *  1: Fewer than the pre-specified fraction of the extracted
              data points are valid.
           *  2: Exceeded maximum number of iterations.
           *  3: Singular matrix in harmonic fit, results may not be
              valid.  This also signals an insufficient number of
              data points to fit.
           *  4: Small or wrong gradient, or ellipse diverged. Subsequent
              ellipses at larger or smaller semimajor axis may have
              the same constant geometric parameters. It's also used
              when the user turns off the fitting algorithm via the
              ``maxrit`` fitting parameter (see the
              `~photutils.isophote.Ellipse` class).
           *  5: Ellipse diverged; not even the minimum number of
              iterations could be executed. Subsequent ellipses at
              larger or smaller semimajor axis may have the same
              constant geometric parameters.
           * -1: Internal use.

    Attributes
    ----------
    rms : float
        The root-mean-square of intensity values along the elliptical
        path.
    int_err : float
        The error of the mean (rms / sqrt(# data points)).
    ellip_err : float
        The ellipticity error.
    pa_err : float
        The position angle error (radians).
    x0_err : float
        The error associated with the center x coordinate.
    y0_err : float
        The error associated with the center y coordinate.
    pix_stddev : float
        The estimate of pixel standard deviation (rms * sqrt(average
        sector integration area)).
    grad : float
        The local radial intensity gradient.
    grad_error : float
        The measurement error of the local radial intensity gradient.
    grad_r_error : float
        The relative error of local radial intensity gradient.
    tflux_e : float
        The sum of all pixels inside the ellipse.
    npix_e : int
        The total number of valid pixels inside the ellipse.
    tflux_c : float
        The sum of all pixels inside a circle with the same ``sma`` as
        the ellipse.
    npix_c : int
        The total number of valid pixels inside a circle with the same
        ``sma`` as the ellipse.
    sarea : float
        The average sector area on the isophote (pixel**2).
    ndata : int
        The number of extracted data points.
    nflag : int
        The number of discarded data points. Data points can be
        discarded either because they are physically outside the image
        frame boundaries, because they were rejected by sigma-clipping,
        or they are masked.
    a3, b3, a4, b4 : float
        The higher order harmonics that measure the deviations from a
        perfect ellipse.  These values are actually the raw harmonic
        amplitudes divided by the local radial gradient and the
        semimajor axis length, so they can directly be compared with
        each other.
    a3_err, b3_err, a4_err, b4_err : float
        The errors associated with the ``a3``, ``b3``, ``a4``, and
        ``b4`` attributes.
    """
    def __init__(self, sample, niter, valid, stop_code):
        self.sample = sample
        self.niter = niter
        self.valid = valid
        self.stop_code = stop_code

        self.intens = sample.mean
        self.rms = np.std(sample.values[2])
        self.int_err = self.rms / np.sqrt(sample.actual_points)
        self.pix_stddev = self.rms * np.sqrt(sample.sector_area)
        self.grad = sample.gradient
        self.grad_error = sample.gradient_error

        self.grad_r_error = sample.gradient_relative_error
        self.sarea = sample.sector_area
        self.ndata = sample.actual_points
        self.nflag = sample.total_points - sample.actual_points

        # flux contained inside ellipse and circle
        (self.tflux_e, self.tflux_c, self.npix_e,
         self.npix_c) = self._compute_fluxes()

        self._compute_errors()

        # deviations from a perfect ellipse
        (self.a3, self.b3, self.a3_err,
         self.b3_err) = self._compute_deviations(sample, 3)
        (self.a4, self.b4, self.a4_err,
         self.b4_err) = self._compute_deviations(sample, 4)

    # This method is useful for sorting lists of instances. Note
    # that __lt__ is the python3 way of supporting sorting.
    def __lt__(self, other):
        if hasattr(other, 'sma'):
            return self.sma < other.sma
        raise ValueError('Comparison object does not have a "sma" attribute.')

    def __str__(self):
        return str(self.to_table())

    @property
    def sma(self):
        """The semimajor axis length (pixels)."""
        return self.sample.geometry.sma

    @property
    def eps(self):
        """The ellipticity of the ellipse."""
        return self.sample.geometry.eps

    @property
    def pa(self):
        """The position angle (radians) of the ellipse."""
        return self.sample.geometry.pa

    @property
    def x0(self):
        """The center x coordinate (pixel)."""
        return self.sample.geometry.x0

    @property
    def y0(self):
        """The center y coordinate (pixel)."""
        return self.sample.geometry.y0

    def _compute_fluxes(self):
        """
        Compute integrated flux inside ellipse, as well as inside a
        circle defined with the same semimajor axis.

        Pixels in a square section enclosing circle are scanned; the
        distance of each pixel to the isophote center is compared both
        with the semimajor axis length and with the length of the
        ellipse radius vector, and integrals are updated if the pixel
        distance is smaller.
        """

        # Compute limits of square array that encloses circle.
        sma = self.sample.geometry.sma
        x0 = self.sample.geometry.x0
        y0 = self.sample.geometry.y0
        xsize = self.sample.image.shape[1]
        ysize = self.sample.image.shape[0]

        imin = max(0, int(x0 - sma - 0.5) - 1)
        jmin = max(0, int(y0 - sma - 0.5) - 1)
        imax = min(xsize, int(x0 + sma + 0.5) + 1)
        jmax = min(ysize, int(y0 + sma + 0.5) + 1)

        # Integrate
        if (jmax-jmin > 1) and (imax-imin) > 1:
            y, x = np.mgrid[jmin:jmax, imin:imax]
            radius, angle = self.sample.geometry.to_polar(x, y)
            radius_e = self.sample.geometry.radius(angle)

            midx = (radius <= sma)
            values = self.sample.image[y[midx], x[midx]]
            tflux_c = np.ma.sum(values)
            npix_c = np.ma.count(values)

            midx2 = (radius <= radius_e)
            values = self.sample.image[y[midx2], x[midx2]]
            tflux_e = np.ma.sum(values)
            npix_e = np.ma.count(values)
        else:
            tflux_e = 0.
            tflux_c = 0.
            npix_e = 0
            npix_c = 0

        return tflux_e, tflux_c, npix_e, npix_c

    def _compute_deviations(self, sample, n):
        """
        Compute deviations from a perfect ellipse, based on the
        amplitudes and errors for harmonic "n". Note that we first
        subtract the first and second harmonics from the raw data.
        """

        try:
            coeffs = fit_first_and_second_harmonics(self.sample.values[0],
                                                    self.sample.values[2])
            coeffs = coeffs[0]
            model = first_and_second_harmonic_function(self.sample.values[0],
                                                       coeffs)
            residual = self.sample.values[2] - model

            c = fit_upper_harmonic(sample.values[0], residual, n)
            covariance = c[1]
            ce = np.diagonal(covariance)
            c = c[0]

            a = c[1] / self.sma / sample.gradient
            b = c[2] / self.sma / sample.gradient

            # this comes from the old code. Likely it was based on
            # empirical experience with the STSDAS task, so we leave
            # it here without too much thought.
            gre = self.grad_r_error if self.grad_r_error is not None else 0.64

            a_err = abs(a) * np.sqrt((ce[1] / c[1])**2 + gre**2)
            b_err = abs(b) * np.sqrt((ce[2] / c[2])**2 + gre**2)

        except Exception:  # we want to catch everything
            a = b = a_err = b_err = None

        return a, b, a_err, b_err

    def _compute_errors(self):
        """
        Compute parameter errors based on the diagonal of the covariance
        matrix of the four harmonic coefficients for harmonics n=1 and
        n=2.
        """

        try:
            coeffs = fit_first_and_second_harmonics(self.sample.values[0],
                                                    self.sample.values[2])
            covariance = coeffs[1]
            coeffs = coeffs[0]
            model = first_and_second_harmonic_function(self.sample.values[0],
                                                       coeffs)
            residual_rms = np.std(self.sample.values[2] - model)
            errors = np.diagonal(covariance) * residual_rms

            eps = self.sample.geometry.eps
            pa = self.sample.geometry.pa

            # parameter errors result from direct projection of
            # coefficient errors.  These showed to be the error estimators
            # that best convey the errors measured in Monte Carlo
            # experiments (see Busko 1996; ASPC 101, 139).
            ea = abs(errors[2] / self.grad)
            eb = abs(errors[1] * (1. - eps) / self.grad)
            self.x0_err = np.sqrt((ea * np.cos(pa))**2 + (eb * np.sin(pa))**2)
            self.y0_err = np.sqrt((ea * np.sin(pa))**2 + (eb * np.cos(pa))**2)
            self.ellip_err = (abs(2. * errors[4] * (1. - eps) / self.sma /
                                  self.grad))
            if abs(eps) > np.finfo(float).resolution:
                self.pa_err = (abs(2. * errors[3] * (1. - eps) / self.sma /
                                   self.grad / (1. - (1. - eps)**2)))
            else:
                self.pa_err = 0.
        except Exception:  # we want to catch everything
            self.x0_err = self.y0_err = self.pa_err = self.ellip_err = 0.

    def fix_geometry(self, isophote):
        """
        Fix the geometry of a problematic isophote to be identical to
        the input isophote.

        This method should be called when the fitting goes berserk and
        delivers an isophote with bad geometry, such as ellipticity > 1
        or another meaningless situation. This is not a problem in
        itself when fitting any given isophote, but will create an error
        when the affected isophote is used as starting guess for the
        next fit.

        Parameters
        ----------
        isophote : `~photutils.isophote.Isophote` instance
            The isophote from which to take the geometry information.
        """

        self.sample.geometry.eps = isophote.sample.geometry.eps
        self.sample.geometry.pa = isophote.sample.geometry.pa
        self.sample.geometry.x0 = isophote.sample.geometry.x0
        self.sample.geometry.y0 = isophote.sample.geometry.y0

    def sampled_coordinates(self):
        """
        Return the (x, y) coordinates where the image was sampled in
        order to get the intensities associated with this isophote.

        Returns
        -------
        x, y : 1D `~numpy.ndarray`
            The x and y coordinates as 1D arrays.
        """

        return self.sample.coordinates()

    def to_table(self):
        """
        Return the main isophote parameters as an astropy
        `~astropy.table.QTable`.

        Returns
        -------
        result : `~astropy.table.QTable`
            An astropy `~astropy.table.QTable` containing the main
            isophote paramters.
        """

        return _isophote_list_to_table([self])


class CentralPixel(Isophote):
    """
    Specialized Isophote class for the galaxy central pixel.

    This class holds only a single intensity value at the central
    position.  Thus, most of its attributes are hardcoded to `None` or a
    default value when appropriate.

    Parameters
    ----------
    sample : `~photutils.utils.EllipseSample` instance
        The sample information.
    """

    def __init__(self, sample):
        self.sample = sample
        self.niter = 0
        self.valid = True
        self.stop_code = 0

        self.intens = sample.mean

        # some values are set to zero to ease certain tasks
        # such as model building and plotting magnitude errors
        self.rms = None
        self.int_err = 0.0
        self.pix_stddev = None
        self.grad = 0.0
        self.grad_error = None
        self.grad_r_error = None
        self.sarea = None
        self.ndata = sample.actual_points
        self.nflag = sample.total_points - sample.actual_points

        self.tflux_e = self.tflux_c = self.npix_e = self.npix_c = None

        self.a3 = self.b3 = 0.0
        self.a4 = self.b4 = 0.0
        self.a3_err = self.b3_err = 0.0
        self.a4_err = self.b4_err = 0.0

        self.ellip_err = 0.
        self.pa_err = 0.
        self.x0_err = 0.
        self.y0_err = 0.

    @property
    def eps(self):
        return 0.

    @property
    def pa(self):
        return 0.

    @property
    def x0(self):
        return self.sample.geometry.x0


class IsophoteList:
    """
    Container class that provides the same attributes as the
    `~photutils.isophote.Isophote` class, but for a list of isophotes.

    The attributes of this class are arrays representing the values of
    the attributes for the entire list of `~photutils.isophote.Isophote`
    instances.  See the `~photutils.isophote.Isophote` class for a
    description of the attributes.

    The class extends the `list` functionality, thus provides basic list
    behavior such as slicing, appending, and support for '+' and '+='
    operators.

    Parameters
    ----------
    iso_list : list of `~photutils.isophote.Isophote`
        A list of `~photutils.isophote.Isophote` instances.
    """

    def __init__(self, iso_list):
        self._list = iso_list

    def __len__(self):
        return len(self._list)

    def __delitem__(self, index):
        self._list.__delitem__(index)

    def __setitem__(self, index, value):
        self._list.__setitem__(index, value)

    def __getitem__(self, index):
        if isinstance(index, slice):
            return IsophoteList(self._list[index])
        return self._list.__getitem__(index)

    def __iter__(self):
        return self._list.__iter__()

    def sort(self):
        self._list.sort()

    def insert(self, index, value):
        self._list.insert(index, value)

    def append(self, value):
        self.insert(len(self) + 1, value)

    def extend(self, value):
        self._list.extend(value._list)

    def __iadd__(self, value):
        self.extend(value)
        return self

    def __add__(self, value):
        temp = self._list[:]  # shallow copy
        temp.extend(value._list)
        return IsophoteList(temp)

    def get_closest(self, sma):
        """
        Return the `~photutils.isophote.Isophote` instance that has the
        closest semimajor axis length to the input semimajor axis.

        Parameters
        ----------
        sma : float
            The semimajor axis length.

        Returns
        -------
        isophote : `~photutils.isophote.Isophote` instance
            The isophote with the closest semimajor axis value.
        """

        index = (np.abs(self.sma - sma)).argmin()
        return self._list[index]

    def _collect_as_array(self, attr_name):
        return np.array(self._collect_as_list(attr_name), dtype=float)

    def _collect_as_list(self, attr_name):
        return [getattr(iso, attr_name) for iso in self._list]

    @property
    def sample(self):
        """
        The isophote `~photutils.isophote.EllipseSample` information.
        """
        return self._collect_as_list('sample')

    @property
    def sma(self):
        """The semimajor axis length (pixels)."""
        return self._collect_as_array('sma')

    @property
    def intens(self):
        """The mean intensity value along the elliptical path."""
        return self._collect_as_array('intens')

    @property
    def int_err(self):
        """The error of the mean intensity (rms / sqrt(# data points))."""
        return self._collect_as_array('int_err')

    @property
    def eps(self):
        """The ellipticity of the ellipse."""
        return self._collect_as_array('eps')

    @property
    def ellip_err(self):
        """The ellipticity error."""
        return self._collect_as_array('ellip_err')

    @property
    def pa(self):
        """The position angle (radians) of the ellipse."""
        return self._collect_as_array('pa')

    @property
    def pa_err(self):
        """The position angle error (radians)."""
        return self._collect_as_array('pa_err')

    @property
    def x0(self):
        """The center x coordinate (pixel)."""
        return self._collect_as_array('x0')

    @property
    def x0_err(self):
        """The error associated with the center x coordinate."""
        return self._collect_as_array('x0_err')

    @property
    def y0(self):
        """The center y coordinate (pixel)."""
        return self._collect_as_array('y0')

    @property
    def y0_err(self):
        """The error associated with the center y coordinate."""
        return self._collect_as_array('y0_err')

    @property
    def rms(self):
        """
        The root-mean-square of intensity values along the elliptical
        path.
        """
        return self._collect_as_array('rms')

    @property
    def pix_stddev(self):
        """
        The estimate of pixel standard deviation (rms * sqrt(average
        sector integration area)).
        """
        return self._collect_as_array('pix_stddev')

    @property
    def grad(self):
        """The local radial intensity gradient."""
        return self._collect_as_array('grad')

    @property
    def grad_error(self):
        """
        The measurement error of the local radial intensity
        gradient.
        """
        return self._collect_as_array('grad_error')

    @property
    def grad_r_error(self):
        """
        The relative error of local radial intensity gradient.
        """
        return self._collect_as_array('grad_r_error')

    @property
    def sarea(self):
        """The average sector area on the isophote (pixel**2)."""
        return self._collect_as_array('sarea')

    @property
    def ndata(self):
        """The number of extracted data points."""
        return self._collect_as_array('ndata')

    @property
    def nflag(self):
        """
        The number of discarded data points. Data points can be
        discarded either because they are physically outside the image
        frame boundaries, because they were rejected by sigma-clipping,
        or they are masked.
        """
        return self._collect_as_array('nflag')

    @property
    def niter(self):
        """The number of iterations used to fit the isophote."""

        return self._collect_as_array('niter')

    @property
    def valid(self):
        """The status of the fitting operation."""
        return self._collect_as_array('valid')

    @property
    def stop_code(self):
        """The fitting stop code."""
        return self._collect_as_array('stop_code')

    @property
    def tflux_e(self):
        """The sum of all pixels inside the ellipse."""
        return self._collect_as_array('tflux_e')

    @property
    def tflux_c(self):
        """
        The sum of all pixels inside a circle with the same ``sma`` as
        the ellipse.
        """
        return self._collect_as_array('tflux_c')

    @property
    def npix_e(self):
        """The total number of valid pixels inside the ellipse."""
        return self._collect_as_array('npix_e')

    @property
    def npix_c(self):
        """
        The total number of valid pixels inside a circle with the same
        ``sma`` as the ellipse.
        """
        return self._collect_as_array('npix_c')

    @property
    def a3(self):
        """
        A third-order harmonic coefficent.  See the
        :func:`~photutils.isophote.fit_upper_harmonic` function for
        details.
        """
        return self._collect_as_array('a3')

    @property
    def b3(self):
        """
        A third-order harmonic coefficent.  See the
        :func:`~photutils.isophote.fit_upper_harmonic` function for
        details.
        """
        return self._collect_as_array('b3')

    @property
    def a4(self):
        """
        A fourth-order harmonic coefficent.  See the
        :func:`~photutils.isophote.fit_upper_harmonic` function for
        details.
        """
        return self._collect_as_array('a4')

    @property
    def b4(self):
        """
        A fourth-order harmonic coefficent.  See the
        :func:`~photutils.isophote.fit_upper_harmonic` function for
        details.
        """
        return self._collect_as_array('b4')

    @property
    def a3_err(self):
        """
        The error associated with `~photutils.isophote.IsophoteList.a3`.
        """
        return self._collect_as_array('a3_err')

    @property
    def b3_err(self):
        """
        The error associated with `~photutils.isophote.IsophoteList.b3`.
        """
        return self._collect_as_array('b3_err')

    @property
    def a4_err(self):
        """
        The error associated with `~photutils.isophote.IsophoteList.a4`.
        """
        return self._collect_as_array('a4_err')

    @property
    def b4_err(self):
        """
        The error associated with `~photutils.isophote.IsophoteList.b3`.
        """
        return self._collect_as_array('b4_err')

    def to_table(self):
        """
        Convert an `~photutils.isophote.IsophoteList` instance to a
        `~astropy.table.QTable` with the main isophote parameters.

        Returns
        -------
        result : `~astropy.table.QTable`
            An astropy QTable with the main isophote parameters.
        """
        return _isophote_list_to_table(self)


def _isophote_list_to_table(isophote_list):
    """
    Convert an `~photutils.isophote.IsophoteList` instance to
    a `~astropy.table.QTable`.

    Parameters
    ----------
    isophote_list : list of `~photutils.isophote.Isophote` or \
            `~photutils.isophote.IsophoteList` instance
        A list of isophotes.

    Returns
    -------
    result : `~astropy.table.QTable`
        An astropy QTable with the main isophote parameters.
    """

    properties = OrderedDict()
    properties['sma'] = 'sma'
    properties['intens'] = 'intens'
    properties['int_err'] = 'intens_err'
    properties['eps'] = 'ellipticity'
    properties['ellip_err'] = 'ellipticity_err'
    properties['pa'] = 'pa'
    properties['pa_err'] = 'pa_err'
    properties['grad'] = 'grad'
    properties['grad_error'] = 'grad_err'
    properties['grad_r_error'] = 'grad_rerr'
    properties['x0'] = 'x0'
    properties['x0_err'] = 'x0_err'
    properties['y0'] = 'y0'
    properties['y0_err'] = 'y0_err'
    properties['ndata'] = 'ndata'
    properties['nflag'] = 'flag'
    properties['niter'] = 'niter'
    properties['stop_code'] = 'stop_code'

    isotable = QTable()

    for k, v in properties.items():
        isotable[v] = np.array([getattr(iso, k) for iso in isophote_list])

        if k in ('pa', 'pa_err'):
            isotable[v] = isotable[v] * 180. / np.pi * u.deg

    return isotable



from matplotlib import pyplot as plt
from astropy.table import Table
from numpy import copy, log10
import tbridge


def adjust_profile(x, y, m_0=27, arcconv=0.168):

    x_new, y_new = copy(x), copy(y)
    y_new = -2.5 * log10(y / (arcconv ** 2)) + m_0

    return x_new, y_new


def single_bin_plot(table_arrays, colours=None, ind_profile_alpha=1, medians=False, to_sb=True,
                    bootstraps=False, ylim=(30, 20),
                    xlabel=None, ylabel=None, output="", dpi=150):

    plt.figure(figsize=(8, 6))

    for i in range(0, len(table_arrays)):
        for j in range(0, len(table_arrays[i])):
            this_prof = table_arrays[i][j]

            sma, intens = this_prof["sma"], this_prof["intens"]
            sma, intens = adjust_profile(sma, intens)

            plt.plot(sma, intens, c=colours[i], alpha=ind_profile_alpha, lw=1)

        if medians:
            interps = tbridge.as_interpolations(table_arrays[i])
            median_sma, median_interp = tbridge.get_median(interps, tbridge.bin_max(table_arrays[i]))
            median_intens = median_interp(median_sma)

            if to_sb:
                median_sma, median_intens = adjust_profile(median_sma, median_intens)

            plt.plot(median_sma, median_intens, lw=3, c=colours[i])

    if ylim is None:
        pass
    else:
        plt.ylim(ylim[0], ylim[1])
    plt.xlim(1,)
    plt.xscale("log")

    if xlabel is not None:
        plt.xlabel(xlabel)
    if ylabel is not None:
        plt.ylabel(ylabel)

    plt.tight_layout()

    if output == "":
        plt.show()
    else:
        plt.savefig(output, dpi=dpi)


def single_prof(x, y, error=None):
    plt.figure(figsize=(8, 6))

    if error is None:
        plt.plot(x, y, color="orange", lw=2)
    else:
        plt.errorbar(x, y, yerr=error)

    plt.tight_layout()
    plt.show()



"""
This module contains the methods required to run KOE, which is a tool designed to automatically extract profiles
from a directory of images, with a given input catalogue.
"""

import galprime
import os

from astropy.io import fits
from astropy.table import Table
from astropy.nddata import Cutout2D
from pebble import ProcessPool

import numpy as np

import warnings
warnings.filterwarnings("ignore")


def process_object(image, image_wcs, object_info, config):
    """ Process a single object, from the whole image to the final profile, for a given location.

    """
    params = {}
    # Generate the cutout
    loc = np.asarray(image_wcs.wcs_world2pix(object_info[config["RA"]],
                                             object_info[config["DEC"]], 0), dtype=int)
    cutout = Cutout2D(image, (loc[0], loc[1]), size=config["SIZE"], wcs=image_wcs).data

    #
    if config["BG_PARAMS"] == '2D':
        sm = galprime.SourceMask(cutout)
        source_mask = sm.multiple(filter_fwhm=[1, 3, 5], tophat_size=[4, 2, 1])
        bg = galprime.background_2D(cutout, source_mask,
                                   box_size=config["BOX_SIZE"], filter_size=config["FILTER_SIZE"])
        params["2D_BG_MED"] = bg.background_median
        params["2D_BG_RMS_MED"] = bg.background_rms_median
        cutout = cutout - bg.background
    # Mask the cutout
    masked_cutout, mask_params = galprime.mask_cutout(cutout, config=config)

    params.update(mask_params)

    for key in object_info.colnames:
        params[key] = object_info[key]

    # Process extraction
    profile = galprime.isophote_fitting(masked_cutout, config=config)

    # Save the data if applicable

    return {"PROFILE": profile.to_table(), "PARAMS": params}


def koe_pipeline(config):
    """ The entire KOE pipeline configured to work with TBriDGE methods.

    """
    # Load in image
    image_filenames = galprime.get_image_filenames(config["IMAGE_DIRECTORY"])[:]

    # Get objects available for profile extraction
    objects = Table.read(config["CATALOG"])

    for image_filename in image_filenames:
        image, wcs = galprime.select_image(image_filename), galprime.get_wcs(image_filename)

        ra_lims, dec_lims = galprime.extraction_limits(wcs, image.shape, config["SIZE"])

        objects_available = galprime.trim_objects(objects, ra_lims, dec_lims, mag_lim=(0, 20),
                                                 ra_key=config["RA"], dec_key=config["DEC"])

        # Prep the container for all objects
        if not os.path.isdir(config["OUT_DIR"]):
            os.mkdir(config["OUT_DIR"])

        out_filename = os.path.splitext(image_filename.split("/")[len(image_filename.split("/")) - 1])[0]
        output = fits.HDUList()

        # Go through all objects and try to extract a valid result. With this method, we can avoid any pesky
        # infinite loops and maximize computation with multithreading.
        max_index = min([len(objects_available), 10]) if config["TEST_MODE"] else len(objects_available)

        job_list = []
        with ProcessPool(max_workers=config["CORES"]) as pool:
            for i in range(max_index):
                job_list.append(pool.schedule(process_object,
                                              args=(image, wcs, objects_available[i], config),
                                              timeout=config["ALARM_TIME"]))
        # Collect the results
        for i in range(len(job_list)):
            try:
                result = job_list[i].result()

                profile = result["PROFILE"]
                if len(profile) == 0:
                    continue
                # Put the profile into a proper saveable format
                valid_colnames = ["sma", "intens", "intens_err", "ellipticity", "ellipticity_err",
                                  "pa", "pa_err", "x0", "x0_err", "y0", "y0_err", "ndata", "flag", "niter"]
                out_table = Table([profile[col] for col in valid_colnames], names=valid_colnames)

                # Generate hdu header
                header = fits.Header()
                for key in result["PARAMS"]:
                    try:
                        header[key] = result["PARAMS"][key]
                    except:
                        continue

                output.append(fits.BinTableHDU(data=out_table,
                                               header=header))

            except Exception as error:
                print(error.args, i)

        output.writeto(config["OUT_DIR"] + out_filename + "_profiles.fits", overwrite=True)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
This module provides a class to sample data along an elliptical path.
"""

import copy

import numpy as np

from .geometry import EllipseGeometry
from .integrator import INTEGRATORS

__all__ = ['EllipseSample']


class EllipseSample:
    """
    Class to sample image data along an elliptical path.

    The image intensities along the elliptical path can be extracted
    using a selection of integration algorithms.

    The ``geometry`` attribute describes the geometry of the elliptical
    path.

    Parameters
    ----------
    image : 2D `~numpy.ndarray`
        The input image.
    sma : float
        The semimajor axis length in pixels.
    x0, y0 : float, optional
        The (x, y) coordinate of the ellipse center.
    astep : float, optional
        The step value for growing/shrinking the semimajor axis. It can
        be expressed either in pixels (when ``linear_growth=True``) or
        as a relative value (when ``linear_growth=False``).  The default
        is 0.1.
    eps : float, optional
        The ellipticity of the ellipse.  The default is 0.2.
    pa : float, optional
        The position angle of ellipse in relation to the positive x axis
        of the image array (rotating towards the positive y axis).  The
        default is 0.
    sclip : float, optional
        The sigma-clip sigma value.  The default is 3.0.
    nclip : int, optional
        The number of sigma-clip interations.  Set to zero to skip
        sigma-clipping.  The default is 0.
    linear_growth : bool, optional
        The semimajor axis growing/shrinking mode.  The default is
        `False`.
    integrmode : {'bilinear', 'nearest_neighbor', 'mean', 'median'}, optional
        The area integration mode.  The default is 'bilinear'.
    geometry : `~photutils.isophote.EllipseGeometry` instance or `None`
        The geometry that describes the ellipse. This can be used in
        lieu of the explicit specification of parameters ``sma``,
        ``x0``, ``y0``, ``eps``, etc.  In any case, the
        `~photutils.isophote.EllipseGeometry` instance becomes an
        attribute of the `~photutils.isophote.EllipseSample` object.
        The default is `None`.

    Attributes
    ----------
    values : 2D `~numpy.ndarray`
        The sampled values as a 2D array, where the rows contain the
        angles, radii, and extracted intensity values, respectively.
    mean : float
        The mean intensity along the elliptical path.
    geometry : `~photutils.isophote.EllipseGeometry` instance
        The geometry of the elliptical path.
    gradient : float
        The local radial intensity gradient.
    gradient_error : float
        The error associated with the local radial intensity gradient.
    gradient_relative_error : float
        The relative error associated with the local radial intensity
        gradient.
    sector_area : float
        The average area of the sectors along the elliptical path from
        which the sample values were integrated.
    total_points : int
        The total number of sample values that would cover the entire
        elliptical path.
    actual_points : int
        The actual number of sample values that were taken from the
        image. It can be smaller than ``total_points`` when the ellipse
        encompasses regions outside the image, or when sigma-clipping
        removed some of the points.
    """

    def __init__(self, image, sma, x0=None, y0=None, astep=0.1, eps=0.2,
                 position_angle=0., sclip=3., nclip=0, linear_growth=False,
                 integrmode='bilinear', geometry=None):
        self.image = image
        self.integrmode = integrmode

        if geometry:
            # when the geometry is inherited from somewhere else,
            # its sma attribute must be replaced by the value
            # explicitly passed to the constructor.
            self.geometry = copy.deepcopy(geometry)
            self.geometry.sma = sma
        else:
            # if no center was specified, assume it's roughly
            # coincident with the image center
            _x0 = x0
            _y0 = y0
            if not _x0 or not _y0:
                _x0 = image.shape[1] / 2
                _y0 = image.shape[0] / 2

            self.geometry = EllipseGeometry(_x0, _y0, sma, eps,
                                            position_angle, astep,
                                            linear_growth)

        # sigma-clip parameters
        self.sclip = sclip
        self.nclip = nclip

        # extracted values associated with this sample.
        self.values = None
        self.mean = None
        self.gradient = None
        self.gradient_error = None
        self.gradient_relative_error = None
        self.sector_area = None

        # total_points reports the total number of pairs angle-radius that
        # were attempted. actual_points reports the actual number of sampled
        # pairs angle-radius that resulted in valid values.
        self.total_points = 0
        self.actual_points = 0

    def extract(self):
        """
        Extract sample data by scanning an elliptical path over the
        image array.

        Returns
        -------
        result : 2D `~numpy.ndarray`
            The rows of the array contain the angles, radii, and
            extracted intensity values, respectively.
        """

        # the sample values themselves are kept cached to prevent
        # multiple calls to the integrator code.
        if self.values is not None:
            return self.values
        else:
            s = self._extract()
            self.values = s
            return s

    def _extract(self, phi_min=0.05):
        # Here the actual sampling takes place. This is called only once
        # during the life of an EllipseSample instance, because it's an
        # expensive calculation. This method should not be called from
        # external code.
        # If one wants to force it to re-run, then do:
        #
        #   sample.values = None
        #
        # before calling sample.extract()

        # individual extracted sample points will be stored in here
        angles = []
        radii = []
        intensities = []
        sector_areas = []

        # reset counters
        self.total_points = 0
        self.actual_points = 0

        # build integrator
        integrator = INTEGRATORS[self.integrmode](self.image, self.geometry,
                                                  angles, radii, intensities)

        # initialize walk along elliptical path
        radius = self.geometry.initial_polar_radius
        phi = self.geometry.initial_polar_angle

        # In case of an area integrator, ask the integrator to deliver a
        # hint of how much area the sectors will have. In case of too
        # small areas, tests showed that the area integrators (mean,
        # median) won't perform properly. In that case, we override the
        # caller's selection and use the bilinear integrator regardless.
        if integrator.is_area():
            integrator.integrate(radius, phi)
            area = integrator.get_sector_area()
            # this integration that just took place messes up with the
            # storage arrays and the constructors. We have to build a new
            # integrator instance from scratch, even if it is the same
            # kind as originally selected by the caller.
            angles = []
            radii = []
            intensities = []
            if area < 1.0:
                integrator = INTEGRATORS['bilinear'](
                    self.image, self.geometry, angles, radii, intensities)
            else:
                integrator = INTEGRATORS[self.integrmode](self.image,
                                                          self.geometry,
                                                          angles, radii,
                                                          intensities)

        # walk along elliptical path, integrating at specified
        # places defined by polar vector. Need to go a bit beyond
        # full circle to ensure full coverage.
        while phi <= np.pi*2. + phi_min:
            # do the integration at phi-radius position, and append
            # results to the angles, radii, and intensities lists.
            integrator.integrate(radius, phi)

            # store sector area locally
            sector_areas.append(integrator.get_sector_area())

            # update total number of points
            self.total_points += 1

            # update angle and radius to be used to define
            # next polar vector along the elliptical path
            phistep_ = integrator.get_polar_angle_step()
            phi += min(phistep_, 0.5)
            radius = self.geometry.radius(phi)

        # average sector area is calculated after the integrator had
        # the opportunity to step over the entire elliptical path.
        self.sector_area = np.nanmean(np.array(sector_areas))

        # apply sigma-clipping.
        angles, radii, intensities = self._sigma_clip(angles, radii, intensities)

        # actual number of sampled points, after sigma-clip removed outliers.
        self.actual_points = len(angles)

        # pack results in 2-d array with NaN values cleaned out
        nan_indices = ~np.isnan(np.array(intensities))

        result = np.array([np.array(angles)[nan_indices], np.array(radii)[nan_indices],
                           np.array(intensities)[nan_indices]])

        return result

    def _sigma_clip(self, angles, radii, intensities):
        if self.nclip > 0:
            for i in range(self.nclip):
                # do not use list.copy()! must be python2-compliant.
                angles, radii, intensities = self._iter_sigma_clip(
                    angles[:], radii[:], intensities[:])

        return np.array(angles), np.array(radii), np.array(intensities)

    def _iter_sigma_clip(self, angles, radii, intensities):
        # Can't use scipy or astropy tools because they use masked arrays.
        # Also, they operate on a single array, and we need to operate on
        # three arrays simultaneously. We need something that physically
        # removes the clipped points from the arrays, since that is what
        # the remaining of the `ellipse` code expects.
        r_angles = []
        r_radii = []
        r_intensities = []

        values = np.array(intensities)
        mean = np.nanmean(values)
        sig = np.nanstd(values)
        lower = mean - self.sclip * sig
        upper = mean + self.sclip * sig

        count = 0
        for k in range(len(intensities)):
            if intensities[k] >= lower and intensities[k] < upper:
                r_angles.append(angles[k])
                r_radii.append(radii[k])
                r_intensities.append(intensities[k])
                count += 1

        return r_angles, r_radii, r_intensities

    def update(self, fixed_parameters):
        """
        Update this `~photutils.isophote.EllipseSample` instance.

        This method calls the
        :meth:`~photutils.isophote.EllipseSample.extract` method to get
        the values that match the current ``geometry`` attribute, and
        then computes the the mean intensity, local gradient, and other
        associated quantities.
        """
        self.geometry.fix = fixed_parameters

        step = self.geometry.astep

        # Update the mean value first, using extraction from main sample.
        s = self.extract()
        self.mean = np.nanmean(s[2])

        # Get sample with same geometry but at a different distance from
        # center. Estimate gradient from there.
        gradient, gradient_error = self._get_gradient(step)

        # Check for meaningful gradient. If no meaningful gradient, try
        # another sample, this time using larger radius. Meaningful
        # gradient means something  shallower, but still close to within
        # a factor 3 from previous gradient estimate. If no previous
        # estimate is available, guess it by adding the error to the
        # current gradient.
        previous_gradient = self.gradient
        if not previous_gradient:
            previous_gradient = gradient + gradient_error

            # solution adopted before 08/12/2019
            # previous_gradient = -0.05  # good enough, based on usage

        if gradient >= (previous_gradient / 3.):  # gradient is negative!
            gradient, gradient_error = self._get_gradient(2 * step)

        # If still no meaningful gradient can be measured, try with
        # previous one, slightly shallower. A factor 0.8 is not too far
        # from what is expected from geometrical sampling steps of 10-20%
        # and a deVaucouleurs law or an exponential disk (at least at its
        # inner parts, r <~ 5 req). Gradient error is meaningless in this
        # case.
        if gradient >= (previous_gradient / 3.):
            gradient = previous_gradient * 0.8
            gradient_error = None

        self.gradient = gradient
        self.gradient_error = gradient_error
        if gradient_error:
            self.gradient_relative_error = gradient_error / np.abs(gradient)
        else:
            self.gradient_relative_error = None

    def _get_gradient(self, step):
        gradient_sma = (1. + step) * self.geometry.sma

        gradient_sample = EllipseSample(
            self.image, gradient_sma, x0=self.geometry.x0,
            y0=self.geometry.y0, astep=self.geometry.astep, sclip=self.sclip,
            nclip=self.nclip, eps=self.geometry.eps,
            position_angle=self.geometry.pa,
            linear_growth=self.geometry.linear_growth,
            integrmode=self.integrmode)

        sg = gradient_sample.extract()
        mean_g = np.nanmean(sg[2])
        gradient = (mean_g - self.mean) / self.geometry.sma / step

        s = self.extract()
        sigma = np.nanstd(s[2])
        sigma_g = np.nanstd(sg[2])

        gradient_error = (np.sqrt(sigma**2 / len(s[2]) +
                                  sigma_g**2 / len(sg[2])) /
                          self.geometry.sma / step)

        return gradient, gradient_error

    def coordinates(self):
        """
        Return the (x, y) coordinates associated with each sampled
        point.

        Returns
        -------
        x, y : 1D `~numpy.ndarray`
            The x and y coordinate arrays.
        """

        angles = self.values[0]
        radii = self.values[1]
        x = np.zeros(len(angles))
        y = np.zeros(len(angles))

        for i in range(len(x)):
            x[i] = (radii[i] * np.cos(angles[i] + self.geometry.pa) +
                    self.geometry.x0)

            y[i] = (radii[i] * np.sin(angles[i] + self.geometry.pa) +
                    self.geometry.y0)

        return x, y


class CentralEllipseSample(EllipseSample):
    """
    An `~photutils.isophote.EllipseSample` subclass designed to handle
    the special case of the central pixel in the galaxy image.
    """

    def update(self, fixed_parameters):
        """
        Update this `~photutils.isophote.EllipseSample` instance with
        the intensity integrated at the (x0, y0) center position using
        bilinear integration. The local gradient is set to `None`.

        'fixed_parameters' is ignored in this subclass.
        """

        s = self.extract()
        self.mean = s[2][0]

        self.gradient = None
        self.gradient_error = None
        self.gradient_relative_error = None

    def _extract(self):
        angles = []
        radii = []
        intensities = []

        integrator = INTEGRATORS['bilinear'](self.image, self.geometry,
                                             angles, radii, intensities)
        integrator.integrate(0.0, 0.0)

        self.total_points = 1
        self.actual_points = 1

        return np.array([np.array(angles), np.array(radii),
                         np.array(intensities)])



from matplotlib import pyplot as plt
from astropy.io import fits
from astropy.table import Table

from scipy.interpolate import interp1d
from numpy import copy, log10, isnan, isinf, round


def matrix_plot(directory, x_bins, y_bins,
                stubs=["bare", "bgadded", "noisy", "bgsub"],
                colours=["red", "blue", "green", "violet"], extra_params="_0.0-0.5"):

    print(x_bins, y_bins)

    n_x, n_y = len(x_bins), len(y_bins)

    fig, ax = plt.subplots(n_x, n_y, sharex=True, sharey=True)
    fig.set_figheight(4)
    fig.set_figwidth(5)
    for stub in stubs:
        filename_prefix = directory + stub + "_profiles/"

        for x in range(0, n_x):
            for y in range(0, n_y):
                # ax[x][y].text(0.5, 0.5, str(round(x_bins[n_x - x - 1], 2)) + "," + str(round(y_bins[y], 2)),
                #               color="red")
                pass
    plt.show()
    return None


def load_median_info(filename):
    bare_hdul = fits.open(filename)
    bare_median, l, u = Table.read(bare_hdul[1]), Table.read(bare_hdul[2]), Table.read(bare_hdul[3])
    bare_median_adj = adjust_profile(bare_median["SMA"], bare_median["INTENS"])
    bare_l_1sig_adj = adjust_profile(l["SMA"], l["INTENS_1SIG"])
    bare_u_1sig_adj = adjust_profile(u["SMA"], u["INTENS_1SIG"])
    bare_l_2sig_adj = adjust_profile(l["SMA"], l["INTENS_2SIG"])
    bare_u_2sig_adj = adjust_profile(u["SMA"], u["INTENS_2SIG"])
    bare_l_3sig_adj = adjust_profile(l["SMA"], l["INTENS_3SIG"])
    bare_u_3sig_adj = adjust_profile(u["SMA"], u["INTENS_3SIG"])
    bare_l_5sig_adj = adjust_profile(l["SMA"], l["INTENS_5SIG"])
    bare_u_5sig_adj = adjust_profile(u["SMA"], u["INTENS_5SIG"])


def adjust_profile(x, y, clean=False, shift=0.00):
    # hlr_index, hlr_value = get_half_light_radius(x, y)
    x_new, y_new = copy(x), copy(y)
    y_new = -2.5 * log10((y + shift) / (0.168 ** 2)) + 27

    if clean:
        x_clean, y_clean = [], []
        for n in range(0, len(x_new)):
            if isnan(y_new[n]) or isinf(y_new[n]):
                continue
            else:
                x_clean.append(x_new[n])
                y_clean.append(y_new[n])
        profile_adj = interp1d(x_clean, y_clean, bounds_error=False, fill_value=max(y_clean))

        return x_new, profile_adj(x_new)

    else:
        return x_new, y_new


""" Binning methods

This module contains all necessary methods for binning catalogues according to various parameters.

"""

from astropy.table import Table
import numpy as np
from matplotlib import pyplot as plt


class BinList:
    """ 
        Container class for bins
    
    """
    
    def __init__(self, bins=None):
        if bins is None:
            bins = []
        elif type(bins) == Bin:
            bins = [bins]
        
        self.bins = bins
        self.min_objects = 4
    
    def rebin(self, index, bounds):
        new_bins = []
        for b in self.bins:            
            new_bins.extend(b.rebin(index, bounds))
        self.bins = new_bins
        
class Bin:
    """ Class for a single bin of information
        :param objects: The catalog rows that belong to the given bin
        :type bin_param_names: array_like
        :param object_column_names: The column names when the objects are sorted column-wise instead of row-wise.
        :type bin_param_names: array_like
        :param bin_params (arr): The values that define the bounds of the bin.
        :type bin_param_names: array_like
        :param bin_param_names (arr): Array of parameter names that the bin was made with.
        :type bin_param_names: array_like
    """
    
    def __init__(self, objects=None, object_column_names = None, bin_params=None):
        if bin_params is None:
            bin_params = []

        if objects is None:
            objects = []
        
        self.objects = objects
        self.object_column_names = object_column_names
        self.bin_params = bin_params
        
    
    def rebin(self, index, bounds):
        """ Rebin a bin according to a set of bounds and a 
            given index
            
            index: the column index to bin by
            bounds: The bounds to place the objects in 
        """
        # Create new bins
        current_bin_params = np.copy(self.bin_params)
        outbins = []

        # Generate bins with the new proper bin parameters
        for i in range(len(bounds) - 1):
            this_bin = Bin(object_column_names=self.object_column_names, 
                           bin_params=current_bin_params)
            this_bin.bin_params = np.append(this_bin.bin_params, 
                                            [np.round(bounds[i], 2), 
                                             np.round(bounds[i + 1], 2)])
            outbins.append(this_bin)
            
        bin_column = np.copy(np.transpose(self.objects)[index])
        for i in range(len(bin_column)):
            for j in range(len(bounds) - 1):
                if bounds[j] < bin_column[i] < bounds[j + 1]:
                    outbins[j].objects.append(self.objects[i])
        return outbins
    
    def columns(self):
        return np.transpose(self.objects)
    
    def param_dict(self):
        param_dict = {}
        columns = self.columns()
        for i in range(len(self.object_column_names)):
            param_dict[self.object_column_names[i]] = columns[i]
        return param_dict
    
    def file_prefix(self):
        prefix = "bin_"
        for i in range(0, len(self.bin_params)):
            if (i + 1) % 2 != 0:
                prefix += str(self.bin_params[i]) + "-"
            else:
                prefix += str(self.bin_params[i]) + "_"
        return prefix
    
    def to_table(self, filename=None, overwrite=False):
        """ Creates a astropy.table.Table object for the objects and names contained
        within the bin.

        :param filename: Filename for saving, defaults to None
        :type filename: str, optional
        :param overwrite: Overwrite filename to save, defaults to False
        :type overwrite: bool, optional
        :return: Table of data parameters and associated 
        :rtype: astropy.table.Table()
        """
        columns = self.columns()
        t_out = Table()
        # Add each column (and associated name) to the table
        for i, name in enumerate(self.object_column_names):
            t_out[name] = columns[i]
        
        # If supplied with a filename, write the Table to a fits table 
        if filename is not None:
            t_out.write(filename, format='fits', overwrite=overwrite)

        return t_out

    
def bin_catalog(config):
    """ Bin the inputted catalog.

    This is a convenience function to get a list of bins based on redshift, star-formation, and mass.

    :param config: Values from input config file or user inputs.
    :type config: dict
    :return: List of bins that have been binned according to the config settings.
    """
    
    catalog = Table.read(config["CATALOG"], format='fits')
    
    mags = np.array(catalog[config["MAG_KEY"]])
    r50s = np.array(catalog[config["R50_KEY"]])
    ns = np.array(catalog[config["N_KEY"]])
    ellips = np.array(catalog[config["ELLIP_KEY"]])

    catalog_objects = [mags, r50s, ns, ellips]
    column_names = ["MAGS", "R50S", "NS", "ELLIPS"]
    bin_bounds = []
    
    # Add our bin stuff in
    for i in range(len(config["BIN_NAMES"])):
        catalog_objects.append(np.array(catalog[config["BIN_NAMES"][i]]))
        column_names.append(config["BIN_NAMES"][i])
        bin_bounds.append(config["BIN_" + str(i)])
        
    rebin_indices = np.arange(4, 4 + len(config["BIN_NAMES"]))
    catalog_objects = np.transpose(np.asarray(catalog_objects))
    
    # Create our initial bin
    init_bin = Bin(objects=catalog_objects, object_column_names=column_names)

    binlist = BinList(bins=[init_bin])
    
    for index in range(len(config["BIN_NAMES"])):
        binlist.rebin(index + 4, config["BIN_" + str(index)])
    
#     check = binlist.rebin(4, config["BIN_0"])
    
#     print("Binning by mass finished", "\n")
#     check = binlist.rebin(5, config["BIN_1"])

    return binlist


def plot_bin_params(b, columns=4, filename=None, dpi=150):
    """ Generate a nice plot of a bin's structural parameters

    :param b: Input bin
    :type b: Bin
    :param columns: Number of columns per row, defaults to 4
    :type columns: int, optional
    :param filename: Output filename, defaults to None
        If none, plt.show() is run instead
    :type filename: str, optional
    :param dpi: DPI for saved image, defaults to 150
    :type dpi: int, optional
    """
    
    data = b.param_dict()
    data_names = [n for n in data]
    rows = int(np.ceil(len(data) / columns))
    
    fig, ax = plt.subplots(rows,columns, facecolor="white")
    fig.set_figwidth(columns * 3)
    fig.set_figheight(rows * 3)
    
    row, column = 0, 0
    
    for i in range(rows * columns):
        row, column = int(np.floor(i / columns)), i % columns
        if i >= len(data_names):
            ax[row][column].remove()
            continue
        name = data_names[i]
        ax[row][column].hist(data[name], histtype='step', color="black",
                            bins=25, lw=4)
        ax[row][column].set_title(name)
    
    for i in range(rows):
        ax[i][0].set_ylabel("Bin Size")
    
    plt.suptitle("Bin: " + str(b.bin_params), fontsize=20)
    plt.tight_layout()
    if filename is not None:
        plt.savefig(filename, dpi=dpi)
    else:
        plt.show()




# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
This module provides tools for pixel integration.
"""

import math

import numpy.ma as ma
import numpy as np

__all__ = ['INTEGRATORS', 'NEAREST_NEIGHBOR', 'BILINEAR', 'MEAN', 'MEDIAN']


# integration modes
NEAREST_NEIGHBOR = 'nearest_neighbor'
BILINEAR = 'bilinear'
MEAN = 'mean'
MEDIAN = 'median'


class _Integrator:
    """
    Base class that supports different kinds of pixel integration methods.

    Parameters
    ----------
    image : 2D `~numpy.ndarray`
         The image array.
    geometry : `~photutils.isophote.EllipseGeometry` instance
        object that encapsulates geometry information about current ellipse
    angles : list
        output list; contains the angle values along the elliptical path
    radii : list
        output list; contains the radius values along the elliptical path
    intensities : list
        output list; contains the extracted intensity values along the
        elliptical path
    """

    def __init__(self, image, geometry, angles, radii, intensities):
        self._image = image
        self._geometry = geometry

        self._angles = angles
        self._radii = radii
        self._intensities = intensities

        # for bounds checking
        self._i_range = range(0, self._image.shape[1] - 1)
        self._j_range = range(0, self._image.shape[0] - 1)

    def integrate(self, radius, phi):
        """
        The three input lists (angles, radii, intensities) are
        appended with one sample point taken from the image by
        a chosen integration method.

        Sub classes should implement the actual integration method.

        Parameters
        ----------
        radius : float
            length of radius vector in pixels
        phi : float
            polar angle of radius vector
        """

        raise NotImplementedError

    def _reset(self):
        """
        Starts the results lists anew.

        This method is for internal use and shouldn't
        be used by external callers.
        """

        self._angles = []
        self._radii = []
        self._intensities = []

    def _store_results(self, phi, radius, sample):
        self._angles.append(phi)
        self._radii.append(radius)
        self._intensities.append(sample)

    def get_polar_angle_step(self):
        """
        Returns the polar angle step used to walk over the
        elliptical path.

        The polar angle step is defined by the actual integrator
        subclass.

        Returns
        -------
        float
           the polar angle step
        """

        raise NotImplementedError

    def get_sector_area(self):
        """
        Returns the area of elliptical sectors where the integration
        takes place.

        This area is defined and managed by the actual integrator
        subclass. Depending on the integrator, the area may be a
        fixed constant, or may change along the elliptical path, so
        it's up to the caller to use this information in a correct way.

        Returns
        -------
        float
           the sector area
        """

        raise NotImplementedError

    def is_area(self):
        """
        Returns the type of the integrator.

        An area integrator gets it's value from operating over a (generally
        variable) number of pixels that define a finite area that  lays
        around the elliptical path, at a certain point on the image defined
        by a polar angle and radius values. A pixel integrator, by contrast,
        integrates over a fixed and normally small area related to a single
        pixel on the image. An example is the bilinear integrator, which
        integrates over a small, fixed, 5-pixel area. This method checks if
        the integrator is of the first type or not.

        Returns
        -------
        boolean
           True if this is an area integrator, False otherwise
        """

        raise NotImplementedError


class _NearestNeighborIntegrator(_Integrator):

    def integrate(self, radius, phi):
        self._r = radius

        # Get image coordinates of (radius, phi) pixel
        i = int(radius * math.cos(phi + self._geometry.pa) +
                self._geometry.x0)
        j = int(radius * math.sin(phi + self._geometry.pa) +
                self._geometry.y0)

        # ignore data point if outside image boundaries
        if (i in self._i_range) and (j in self._j_range):
            sample = self._image[j][i]

            if sample is not ma.masked:
                self._store_results(phi, radius, sample)

    def get_polar_angle_step(self):
        return 1. / self._r

    def get_sector_area(self):
        return 1.

    def is_area(self):
        return False


class _BiLinearIntegrator(_Integrator):

    def integrate(self, radius, phi):
        self._r = radius

        # Get image coordinates of (radius, phi) pixel
        x_ = radius * math.cos(phi + self._geometry.pa) + self._geometry.x0
        y_ = radius * math.sin(phi + self._geometry.pa) + self._geometry.y0
        i = int(x_)
        j = int(y_)
        fx = x_ - i
        fy = y_ - j

        # ignore data point if outside image boundaries
        if (i in self._i_range) and (j in self._j_range):
            # in the future, will need to handle masked pixels here
            qx = 1. - fx
            qy = 1. - fy

            if (self._image[j][i] is not ma.masked and
                    self._image[j+1][i] is not ma.masked and
                    self._image[j][i+1] is not ma.masked and
                    self._image[j+1][i+1] is not ma.masked):

                sample = (self._image[j][i] * qx * qy +
                          self._image[j + 1][i] * qx * fy +
                          self._image[j][i + 1] * fx * qy +
                          self._image[j + 1][i + 1] * fy * fx)

                self._store_results(phi, radius, sample)

    def get_polar_angle_step(self):
        return 1. / self._r

    def get_sector_area(self):
        return 2.

    def is_area(self):
        return False


class _AreaIntegrator(_Integrator):

    def __init__(self, image, geometry, angles, radii, intensities):
        super().__init__(image, geometry, angles, radii, intensities)

        # build auxiliary bilinear integrator to be used when
        # sector areas contain a too small number of valid pixels.
        self._bilinear_integrator = INTEGRATORS[BILINEAR](image, geometry,
                                                          angles, radii,
                                                          intensities)

    def integrate(self, radius, phi):
        self._phi = phi

        # Get image coordinates of the four vertices of the elliptical sector.
        vertex_x, vertex_y = self._geometry.initialize_sector_geometry(phi)

        self._sector_area = self._geometry.sector_area

        # step in polar angle to be used by caller next time
        # when updating the current polar angle `phi` to point
        # to the next sector.
        self._phistep = self._geometry.sector_angular_width

        # define rectangular image area that encompasses the elliptical
        # sector. We have to account for rounding of pixel indices.
        i1 = int(min(vertex_x)) - 1
        j1 = int(min(vertex_y)) - 1
        i2 = int(max(vertex_x)) + 1
        j2 = int(max(vertex_y)) + 1

        # polar angle limits for this sector
        phi1, phi2 = self._geometry.polar_angle_sector_limits()

        # ignore data point if the elliptical sector lies
        # partially, ou totally, outside image boundaries
        if (i1 in self._i_range) and (j1 in self._j_range) and \
           (i2 in self._i_range) and (j2 in self._j_range):

            # Scan rectangular image area, compute sample value.
            npix = 0
            accumulator = self.initialize_accumulator()
            for j in range(j1, j2):
                for i in range(i1, i2):
                    # Check if polar coordinates of each pixel
                    # put it inside elliptical sector.
                    rp, phip = self._geometry.to_polar(i, j)

                    # check if inside angular limits
                    if phip < phi2 and phip >= phi1:

                        # check if radius is inside bounding ellipses
                        sma1, sma2 = self._geometry.bounding_ellipses()
                        aux = ((1. - self._geometry.eps) /
                               math.sqrt(((1. - self._geometry.eps) *
                                          math.cos(phip))**2 +
                                         (math.sin(phip))**2))

                        r1 = sma1 * aux
                        r2 = sma2 * aux

                        if rp < r2 and rp >= r1:
                            # update accumulator with pixel value
                            pix_value = self._image[j][i]
                            if pix_value is not ma.masked:
                                accumulator, npix = self.accumulate(
                                    pix_value, accumulator)

            # If 6 or less pixels were sampled, get the bilinear
            # interpolated value instead.
            if npix in range(0, 7):
                # must reset integrator to remove older samples.
                self._bilinear_integrator._reset()
                self._bilinear_integrator.integrate(radius, phi)
                # because it was reset, current value is the only one stored
                # internally in the bilinear integrator instance. Move it
                # from the internal integrator to this instance.
                if len(self._bilinear_integrator._intensities) > 0:
                    sample_value = self._bilinear_integrator._intensities[0]
                    self._store_results(phi, radius, sample_value)

            elif npix > 6:
                sample_value = self.compute_sample_value(accumulator)
                self._store_results(phi, radius, sample_value)

    def get_polar_angle_step(self):
        _, phi2 = self._geometry.polar_angle_sector_limits()
        phistep = self._geometry.sector_angular_width / 2. + phi2 - self._phi
        return phistep

    def get_sector_area(self):
        return self._sector_area

    def is_area(self):
        return True

    def initialize_accumulator(self):
        raise NotImplementedError

    def accumulate(self, pixel_value, accumulator):
        raise NotImplementedError

    def compute_sample_value(self, accumulator):
        raise NotImplementedError


class _MeanIntegrator(_AreaIntegrator):

    def initialize_accumulator(self):
        accumulator = 0.
        self._npix = 0
        return accumulator

    def accumulate(self, pixel_value, accumulator):
        if not np.isnan(pixel_value):
            accumulator += pixel_value
            self._npix += 1
        else:
            pass
        return accumulator, self._npix

    def compute_sample_value(self, accumulator):
        return accumulator / self._npix


class _MedianIntegrator(_AreaIntegrator):

    def initialize_accumulator(self):
        accumulator = []
        self._npix = 0
        return accumulator

    def accumulate(self, pixel_value, accumulator):
        accumulator.append(pixel_value)
        self._npix += 1
        return accumulator, self._npix

    def compute_sample_value(self, accumulator):
        accumulator.sort()
        return accumulator[int(self._npix/2)]


# Specific integrator subclasses can be instantiated from here.
INTEGRATORS = {
    NEAREST_NEIGHBOR: _NearestNeighborIntegrator,
    BILINEAR: _BiLinearIntegrator,
    MEAN: _MeanIntegrator,
    MEDIAN: _MedianIntegrator
}



""" Scripts for generating galaxy models """

from scipy.stats import gaussian_kde
from scipy.special import gamma, gammainc, gammaincinv, kn
from numpy import pi, exp, mgrid, array
from numpy.random import uniform

from astropy.modeling.models import Sersic2D

import galprime

def object_kde(columns):
    return gaussian_kde(columns)


def i_at_r50(mag, n=2, r_50=2, m_0=27):
    """ Get the intensity at the half-light radius """
    b_n = b(n)
    l_tot = 10 ** ((mag - m_0) / -2.5) * (b_n ** (2 * n))
    denom = (r_50 ** 2) * 2 * pi * n * exp(b_n) * gamma(2 * n)
    i_e = l_tot / denom

    return i_e

def b(n, estimate=False):
    """ Get the b_n normalization constant for the sersic profile. 
    From Graham and Driver.
    """
    if estimate:
        return 2 * n - (1 / 3) + (4 / (405 * n)) + (46 / (25515 * (n ** 2)))
    else:
        return gammaincinv(2 * n, 0.5)


def model_from_kde(kde, config=None, mag_kde=None, names=None, seed=None):
    
    cutout_size = 101 if config is None else config["SIZE"]
    arc_conv = 1 if config is None else config["ARC_CONV"]
    
    attempt = array(kde.resample(1, seed=seed), dtype=float)
    params = galprime.ordered_dict(names, attempt)
    
    if mag_kde is not None:
        params["MAGS"] = mag_kde.resample(1, seed=seed)[0]
    
    r_50_pix = params["R50S"] / arc_conv
    
    i_R50 = i_at_r50(params["MAGS"], n=params["NS"], r_50=r_50_pix, m_0=config["ZEROPOINT"])
    theta = uniform(0, 2*pi)
    
    params["I_R50"] = i_R50
    params["PA"] = theta
    params["R50_PIX"] = params["R50S"] / arc_conv
    
    x,y = mgrid[:cutout_size, :cutout_size]
    # Generate the Sersic model from the parameters
    sersic_model = gen_sersic_model(i_R50=i_R50, r_eff=params["R50S"], n=params["NS"], ellip=params["ELLIPS"], 
                                    theta=theta, x_0=cutout_size / 2, y_0=cutout_size/2, 
                                    arc_conv=arc_conv, shape=(cutout_size, cutout_size))
    
    for n in params:
        params[n] = float(params[n])


    return sersic_model, params
    

def gen_sersic_model(i_R50=10, r_eff=5, n=1, ellip=0.1, theta=0, x_0=0, y_0=0, arc_conv=1, shape=(51,51)):
    try:
        sersic_model = Sersic2D(amplitude=i_R50, r_eff=r_eff / arc_conv, n=n, ellip=ellip, theta=theta, x_0=x_0, y_0=y_0)
        x,y = mgrid[:shape[0], :shape[1]]
        z = sersic_model(x,y)

    except ValueError:
        attempt = [i_R50, r_eff, n, ellip, theta, x_0, y_0, arc_conv, shape]
        raise galprime.GalPrimeError("Failed to generate sersic model with the following parameters:\n" + str(attempt))
    
    return z


def check_sersic_params(i_r50, i_r50_max=200):
    pass



from numpy import ceil, sqrt, log
from matplotlib import pyplot as plt


def view_cutouts(cutouts, output="", dpi=150, show_index=False, log_scale=True, limits=(1, -3), cmap="magma_r"):
    """ Generate cutouts plot for input filename"""

    width = int(ceil(sqrt(len(cutouts))))

    print(len(cutouts), "cutouts to visualize. Generating", str(width) + "x" + str(width), "figure.")

    fig, ax = plt.subplots(width, width)

    fig.set_figheight(width)
    fig.set_figwidth(width)

    index = 0
    for x in range(0, width):
        for y in range(0, width):
            ax[x][y].set_xticks([])
            ax[x][y].set_yticks([])
            try:
                image = cutouts[index]
                if log_scale:
                    ax[x][y].imshow(log(image), cmap=cmap, vmax=limits[0], vmin=limits[1])
                else:
                    ax[x][y].imshow(image, cmap=cmap, vmax=limits[0], vmin=limits[1])

                if show_index:
                    ax[x][y].text(5, 0, str(index), color="red", fontweight="bold",
                                  fontsize=10, alpha=0.7, **{'fontname': 'Helvetica'})
            except:
                pass

            index += 1
    plt.tight_layout()
    plt.subplots_adjust(hspace=0.2, wspace=0.2)

    if output == "":
        plt.show()
    else:
        print("Saving figure to " + output)
        plt.savefig(output, dpi=dpi)




""" Classes and methods for handling astronomical data """

import math

from astropy.io import fits
from astropy.table import Table
from astropy import wcs
from astropy.nddata import Cutout2D

from pathlib import Path
import numpy as np

import galprime

from tqdm import tqdm


class CutoutList:
    
    def __init__(self, cutouts=None, cutout_data=[], metadata={}):
        if cutouts is None:
            cutouts = []
        if metadata == None:
            metadata = {}
        
        self.cutouts = cutouts              # List to hold cutouts
        self.cutout_data = cutout_data      # This is the container for associated data to each cutout
        self.metadata = metadata            # Any extra needed metadata
    
        metadata["N_CUTOUTS"] = len(cutouts)
    
    def update_medatada(self):
        self.metadata["N_CUTOUTS"] = len(self.cutouts)
        
    def cutouts_to_file(self, filename="cutouts.fits", overwrite=False):
        
        if len(self.cutouts) == 0:
            return

        out_hdulist = fits.HDUList()
        head = fits.Header()
        
        # Add the metadata (if any) to the header
        for n in self.metadata:
            head[n] = self.metadata[n]

        # Now add the individual cutouts to the HDUList with associated data in the individual headers
        for i, cutout in enumerate(self.cutouts):
            this_head = head.copy()
            try:
                this_head.update(self.cutout_data[i])
            except ValueError:
                pass
            out_hdulist.append(fits.ImageHDU(data=cutout, header=this_head))
        
        out_hdulist.writeto(filename, overwrite=overwrite)



def get_image_filenames(images_directory):
    """
    Retrieves a list of all available filenames for a given directory, and a given band.

    :param images_directory: The top-level directory from which to get image filenames from.
    :type images_directory: str
    :param image_band: The image band, defaults to "i"
    :type image_band: str, optional
    :param check_band: Check to make sure the images are of a certain band, defaults to False.
        WARNING: Optimized for HSC filename format (ex: HSC-I_9813_4c3.fits).
    :type check_band: bool, optional
    """
    image_filenames = []
    images = Path(images_directory).rglob('*.fits')
    for image in images:
        image_filenames.append(str(image))
    return image_filenames


def get_closest_psf(psfs, ra, dec, ra_key="RA", dec_key="DEC"):
    ras = np.array([psfs.cutout_data[i]["RA"] for i in range(len(psfs.cutout_data))])
    decs = np.array([psfs.cutout_data[i]["DEC"] for i in range(len(psfs.cutout_data))])
    
    closest = np.argmin(((ras - ra) ** 2 + (decs - dec) ** 2))
        
    return psfs.cutouts[closest]


def get_wcs(fits_filename):
    """
        Finds and returns the WCS for an image. If Primary Header WCS no good, searches each index until a good one
        is found. If none found, raises a ValueError

        :param fits_filename: The filename for the FITS file.
        :type fits_filename: str
        :return: The WCS object.
        :raises: ValueError
    """
    # Try just opening the initial header
    wcs_init = wcs.WCS(fits_filename)
    ra, dec = wcs_init.axis_type_names
    if ra.upper() == "RA" and dec.upper() == "DEC":
        return wcs_init

    else:
        hdu_list = fits.open(fits_filename)
        for n in hdu_list:
            try:
                wcs_slice = wcs.WCS(n.header)
                ra, dec = wcs_slice.axis_type_names
                if ra.upper() == "RA" and dec.upper() == "DEC":
                    return wcs_slice
            except:
                continue
        hdu_list.close()

    raise ValueError


def gen_backgrounds(image_directory, config=None):
    n_cutouts = 10 if config is None else config["N_BACKGROUNDS"]
    
    image_files = get_image_filenames(image_directory)

    print(image_directory, image_files, n_cutouts)


def get_angular_size_dist(z, H0=71, WM=0.27):
    """
    Return the angular size distance in Megaparsecs.
    (Stripped down version of Cosmocalc by Ned Wright and Tom Aldcroft (aldcroft@head.cfa.harvard.edu))

    :param z: The redshift.
    :type z: float
    :param H0: The Hubble constant, defaults to 71 km/s/Mpc
    :type H0: float, optional
    :param WM: matter density parameter, defaults to 0.27
    :type WM: float, optional
    """
    try:
        c = 299792.458  # velocity of light in km/sec

        if z > 100:
            z /= 299792.458  # Values over 100 are in km/s

        WV = 1.0 - WM - 0.4165 / (H0 * H0)  # Omega(vacuum) or lambda
        age = 0.0  # age of Universe in units of 1/H0

        h = H0 / 100.
        WR = 4.165E-5 / (h * h)  # includes 3 massless neutrino species, T0 = 2.72528
        WK = 1 - WM - WR - WV
        az = 1.0 / (1 + 1.0 * z)
        n = 1000  # number of points in integrals
        for i in range(n):
            a = az * (i + 0.5) / n
            adot = math.sqrt(WK + (WM / a) + (WR / (a * a)) + (WV * a * a))
            age += 1. / adot

        DCMR = 0.0

        # do integral over a=1/(1+z) from az to 1 in n steps, midpoint rule
        for i in range(n):
            a = az + (1 - az) * (i + 0.5) / n
            adot = math.sqrt(WK + (WM / a) + (WR / (a * a)) + (WV * a * a))
            DCMR = DCMR + 1. / (a * adot)

        DCMR = (1. - az) * DCMR / n

        # tangential comoving distance
        ratio = 1.0
        x = math.sqrt(abs(WK)) * DCMR
        if x > 0.1:
            if WK > 0:
                ratio = 0.5 * (math.exp(x) - math.exp(-x)) / x
            else:
                ratio = math.math.sin(x) / x
        else:
            y = x * x
            if WK < 0:
                y = -y
            ratio = 1. + y / 6. + y * y / 120.
        DCMT = ratio * DCMR
        DA = az * DCMT
        Mpc = lambda x: c / H0 * x
        DA_Mpc = Mpc(DA)

        return DA_Mpc
    except:
        raise ValueError


def get_image_filenames(images_directory):
    """
    Retrieves a list of all available filenames for a given directory, and a given band.

    :param images_directory: The top-level directory from which to get image filenames from.
    :type images_directory: str
    :param image_band: The image band, defaults to "i"
    :type image_band: str, optional
    :param check_band: Check to make sure the images are of a certain band, defaults to False.
        WARNING: Optimized for HSC filename format (ex: HSC-I_9813_4c3.fits).
    :type check_band: bool, optional
    """
    image_filenames = []
    images = Path(images_directory).rglob('*.fits')
    for image in images:
        image_filenames.append(str(image))
    return image_filenames


def gen_cutouts(image_directory, config=None, buffer=10, progress_bar=False):
    n_cutouts = 10 if config is None else config["N_BACKGROUNDS"]
    cutout_size = 71 if config is None else config["SIZE"]
    if cutout_size % 2 == 0:
        cutout_size += 1 
    
    hw = cutout_size / 2
    
    image_files = get_image_filenames(image_directory)
    
    cutouts, cutout_data = [], []
    iterator = tqdm(range(n_cutouts), desc="Making Image Cutouts: ") if progress_bar else range(n_cutouts)
    for i in iterator:
        this_fn = np.random.choice(image_files)
        
        with fits.open(this_fn) as HDUList:
            for i in range(len(HDUList)):
                data = HDUList[i].data
                if type(data) == np.ndarray:
                    this_wcs = wcs.WCS(HDUList[i].header)
                    break
        x, y = np.random.randint(hw + buffer, data.shape[0] - hw - buffer), np.random.randint(hw + buffer, data.shape[1] - hw - buffer)
        ra, dec = this_wcs.wcs_pix2world(x,y,0)
        cutout = Cutout2D(data=data, position=[x,y], size=cutout_size, wcs=this_wcs)
        
        cutouts.append(cutout.data)
        
        cutout_data.append({"IMG":this_fn, "X":x, "Y":y, "RA":float(ra), "DEC":float(dec) })
    
    return CutoutList(cutouts=cutouts, cutout_data=cutout_data)


def cutouts_from_file(filename):
    cutouts, cutout_datasets = [], []
    with fits.open(filename) as HDUList:
        for hdu in HDUList:
            try:
                cutout, cutout_data = hdu.data, hdu.header
                cutouts.append(cutout)
                cutout_datasets.append(cutout_data)
            except:
                continue
        
    return CutoutList(cutouts=cutouts, cutout_data=cutout_datasets)


def ordered_dict(names, params):
    out_dict = {}
    for i, n in enumerate(names):
        out_dict[n] = params[i]
    return out_dict


def dict_extend(d, extension="_1"):
    """ Adds an extension onto the dict keys for a given input dictionary.
    Helpful when needing to add multiple dicts to a given metadata file

    :param d: An input dictionary
    :type d: dict
    :param extension: The extension to add to each key, defaults to "_1"
    :type extension: str, optional
    :return: Dictionary with added extensions onto each key
    :rtype: dict
    """
    d_new = {}
    for n in d:
        d_new[n + extension] = d[n]
    return d_new


def good_isolist_names():
     return ['sma', 'intens', 'intens_err', 'ellipticity', 'ellipticity_err', 
              'pa', 'pa_err', 'x0', 'x0_err', 'y0', 'y0_err']


def table_from_isolist(isolist, hdu = True, names=None):
    """ Generates an astropy table

    :param isolist: The photutils.isophote.isophotelist object 
    :type isolist: photutils.isophote.isophotelist
    :param names: Column names, defaults to galprime.good_isolist_names()
    :type names: list, optional
    :return: The table generated from the isolist
    :rtype: astropy.table.Table
    """

    if names is None:
        names = good_isolist_names()
    isolist_t = isolist.to_table()

    return Table([isolist_t[col] for col in names],  names=names)




from distutils.core import setup
from setuptools import find_packages
import galprime

setup(
    name='galprime',
    packages=find_packages(),
    version=galprime.__version__,
    license='BSD-3-Clause',
    description = 'Simulation suite for efficiency testing of galaxy surface brightness profiles.',
    author="Harrison Souchereau",
    author_email='harrison.souchereau@yale.edu',
    url='https://github.com/HSouch/GalPRIME',
    keywords='galaxies surface brightness profiles',
    scripts=[
        # "bin/koe.py",
        # "bin/GalPRIME.py"
    ],
    install_requires=[
        'scikit-image',
        'numpy',
        'photutils',
        'astropy',
        'pebble',
        'tqdm'
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
)


# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
This subpackage contains tools for fitting elliptical isophotes to
galaxy images.

Forked and altered from the photutils.isophote package. It has been
reconfigured to automatically handle nan pixels in masked images.

https://photutils.readthedocs.io/en/stable/

https://photutils.readthedocs.io/en/stable/isophote.html
"""

from .ellipse import *  # noqa
from .fitter import *  # noqa
from .geometry import *  # noqa
from .harmonics import *  # noqa
from .integrator import *  # noqa
from .isophote import *  # noqa
from .model import *  # noqa
from .sample import *  # noqa




__version__ = "2.0.2"

from .backgrounds import *
from .binning import *
from .config import *
from .data import *
from .extraction import *
from .koe import *
from .masking import *
from .modelling import *
from .sims import *

# from .densities import *
# from .medians import *
# from .profiles import *

from . import plotting

def __main__():
    pass


class GalPrimeError(Exception):
    pass


